
; *
; *  field31 can/ip
; *

.include "atmega32.inc"

.include "macro.inc"

.equ	nic_dat	=	0x100	; one page network buffer
.equ	sns_dat	=	0x200	; sensor data
.equ	trg_dat	=	0x240	; trigger values
.equ	can_dat	=	0x280	; can buffer
.equ	udp_sav	=	0x2a0	; udp connection data
.equ	tcp_sav	=	0x300	; tcp connection data
.equ	sns_ent	=	0x400	; can remote sensor buffer

.def	rex	=	r15
.def	msr_tim	=	r14

.def	dl	=	r21
.def	dh	=	r22

; hardware nic address
.equ	MAC0	=	0x00
.equ	MAC1	=	0x01
.equ	MAC2	=	0x82
.equ	MAC3	=	0x5c
.equ	MAC4	=	0x4a
.equ	MAC5	=	0xdb

; local ip address
.equ	IP0	=	192
.equ	IP1	=	168
.equ	IP2	=	13
.equ	IP3	=	13

; one sensor monitoring battery voltage
.equ	sns_num	=	1
; some controling switches
.equ	trg_num	=	4

; can defs
.equ	CAN_CS	=	PB4
.equ	can_addr	=	0x02
.equ	can_nam0	=	'w'
.equ	can_nam1	=	'w'
.equ	can_nam2	=	'w'
.equ	can_nam3	=	'p'
.equ	can_nam4	=	'r'
.equ	can_nam5	=	'o'
.equ	can_nam6	=	'x'
.equ	can_nam7	=	'y'


; macros for funny yellow status led
.macro	led_on
	cbi	PORTD,	PD1
.endmacro
.macro	led_off
	sbi	PORTD,	PD1
.endmacro


.org	0x0000
	rjmp	init
	
.org	TOF1addr
	rjmp	timer_tcp
	
.org	TOF0addr
	rjmp	timer_msr
	
.org	ADCaddr
	reti
	
	
timer_msr:
	in	rex,	SREG

	; delay measurements
	sbrs	msr_tim,	6
	inc	msr_tim
	
	out	SREG,	rex
	reti
	
timer_tcp:
	in	rex,	SREG

	; tcp connection timing
	rcall	tcp_timer

	out	SREG,	rex
	reti
	

init:
	; init stack pointer
	ldi	r16,	LOW( RAMEND )
	out	SPL,	r16
	ldi	r16,	HIGH( RAMEND )
	out	SPH,	r16
	
	; init ports
	clr	r16
	out	PORTA,	r16
	out	DDRA,	r16
	out	PORTB,	r16
	out	DDRB,	r16
	out	PORTC,	r16
	out	DDRC,	r16
	out	PORTD,	r16
	out	DDRD,	r16
	
	sbi	DDRD,	PD1
	sbi	PORTD,	PD0
	
	; init measure timer
	ldi	r16,	( 1 << CS02 ) | ( 1 << CS00 )
	out	TCCR0,	r16
	clr	msr_tim
	
	; init tcp timer
	clr	r16
	out	TCCR1A,	r16
	ldi	r16,	( 1 << CS11 ) | ( 1 << CS10 )
	out	TCCR1B,	r16
	
	ldi	r16,	( 1 << TOIE0 ) | ( 1 << TOIE1 )
	out	TIMSK,	r16
	
	; init inet variables
	call	inet_init
	
	; setup can controler
	rcall	can_init
	
	; setup triggers
	rcall	trg_load
	
	; setup sensor
	rcall	msr_init
	
	; switch on the nic trigger
	clr	dh
	set
	rcall	trg_set
	
	; clear remote sensor cache
	sety	sns_ent
	clr	r16
init_sns_ent:
	st	Y+,	r16
	cpi	YH,	HIGH( sns_ent ) + 1
	brne	init_sns_ent
	
	sei
	
	led_off
	
	; main loop
main_nic:
	clr	dh
	rcall	trg_get
	brtc	main_can
	
	; do internet stuff
	call	inet
	
main_can:
	rcall	can_rcv
	breq	main_msr
	
	; process can message
	led_on
	rcall	can_snd_udp	; can over ether
	call	can_entry	; cache sensor values
	rcall	can_pro	; process commandos
	led_off

main_msr:	
	sbrs	msr_tim,	6
	rjmp	main_nic
	
	clr	msr_tim

	; check push button	
	led_on
	ldi	dh,	3
	clt
	sbic	PIND,	PD0
	set
	rcall	trg_set
	
	; measure supply voltage
	rcall	msr
	
	lds	dl,	sns_dat+sns_flg
	; voltage low ?
main_msr_tst_min:
	sbrs	dl,	sns_flg_aln
	rjmp	main_msr_tst_max
	
	; auto off trigger set ?
	ldi	dh,	2
	rcall	trg_get
	brtc	main_msr_fin
	
	; switch off nic to save battery
	clr	dh
	clt
	rcall	trg_set
	
main_msr_tst_max:
	sbrs	dl,	sns_flg_alx
	rjmp	main_msr_fin
	
	ldi	dh,	1
	rcall	trg_get
	brtc	main_msr_fin
	
	; power good : all systems on 
	clr	dh
	set
	rcall	trg_set
	
main_msr_fin:
	led_off
	rjmp	main_nic


; *
; *  send can message through ethernet interface

can_snd_udp:
	clr	dh
	rcall	trg_get
	brts	can_snd_udp_do_snd
	ret
can_snd_udp_do_snd:
	sety	can_dat
	setz	nic_dat
	
	clr	dh
	ldi	dl,	( ( UDP_DAT - IP_DAT ) + 1 + 0x0d )
	std	Z+UDP_LEN+0,	dh
	std	Z+UDP_LEN+1,	dl
	adiw	ZL,	UDP_DAT
	
	ldi	dl,	udp_cmd_can
	st	Z+,	dl
	
	ldi	dh,	0x0d
can_snd_udp_cpy:
	ld	dl,	Y+
	st	Z+,	dl
	dec	dh
	brne	can_snd_udp_cpy
	
	rjmp	udp_snd_cmd


; *
; *  process can messages	

can_pro:
	sety	can_dat
	
	; verify address
	ldd	dl,	Y+0
	cpi	dl,	can_addr_all
	breq	can_pro_adr_ok
	cpi	dl,	can_addr
	breq	can_pro_adr_ok
can_pro_done:
	ret
	
can_pro_adr_ok:
	; remote frame request (RTR) bit set?
	ldd	dl,	Y+4
	sbrs	dl,	6
	rjmp	can_pro_cmd_trg_set_chk
	
	; yes - check commando code
	ldd	dh,	Y+2
	mov	dl,	dh
	
	cpi	dh,	cmd_prp
	breq	can_pro_cmd_prp
	
	subi	dh,	cmd_sns_0
	cpi	dh,	0x10
	brcs	can_pro_cmd_sns

	mov	dh,	dl	
	subi	dh,	cmd_tr0
	cpi	dh,	trg_num
	brcs	can_pro_cmd_trg

	ret

can_pro_cmd_prp:
	; propagate station id
	setz	( can_msg_prp * 2 )
	sety	can_dat
	ldi	dh,	0x0d
can_pro_cmd_prp1:
	lpm	dl,	Z+
	st	Y+,	dl
	dec	dh
	brne	can_pro_cmd_prp1
	
	; on both interfaces
	sety	can_dat
	rcall	can_snd_y
	rjmp	can_snd_udp
	
can_pro_cmd_sns:
	; respond to sensor query
	clr	dl	; clear event flag
	call	msr_can_snd
	rjmp	can_snd_udp
	
can_pro_cmd_trg:
	; respond to trigger query
	rcall	trg_can_snd
	rjmp	can_snd_udp

can_pro_cmd_trg_set_chk:
	; RTR clear - check for trigger set command
	ldd	dh,	Y+2
	subi	dh,	cmd_tr0s
	cpi	dh,	trg_num
	brcs	can_pro_cmd_trg_set
	ret
can_pro_cmd_trg_set:
	; check if remote triggering allowed
	rcall	trg_ld
	sbrs	dl,	trg_flg_art
	ret
	
	; update trigger
	ldd	dl,	Y+3
	bst	dl,	trg_flg
	rjmp	trg_set


; *
; *  cache sensor values from other can station for html display

can_entry:
	sety	can_dat
	ldd	dh,	Y
	cpi	dh,	can_addr_all
	breq	can_entry_ret
	ldd	dl,	Y+4
	sbrc	dl,	6
can_entry_ret:
	ret
	ldd	dl,	Y+2
	subi	dl,	cmd_sns_0
	cpi	dl,	0x10
	brcs	can_sns_entry
	ret
	
can_sns_entry:
	setz	sns_ent
can_sns_entry_fnd:
	ldd	r16,	Z+0x0e
	tst	r16
	breq	can_sns_entry_match
	cpse	dh,	r16
	rjmp	can_sns_entry_nxt
	ldd	r17,	Z+0x0f
	cpse	dl,	r17
	rjmp	can_sns_entry_nxt
can_sns_entry_match:
	std	Z+0x0e,	dh
	std	Z+0x0f,	dl
	sbrc	dl,	3
	adiw	ZL,	8
	adiw	YL,	5
	ldi	r16,	6
can_sns_entry_cpy:
	ld	r17,	Y+
	st	Z+,	r17
	dec	r16
	brne	can_sns_entry_cpy
	ret
can_sns_entry_nxt:
	adiw	ZL,	16
	cpi	ZH,	HIGH( sns_ent ) + 1
	brne	can_sns_entry_fnd
	ret


.include "eeprom.asm"
.include "trigger.asm"
.include "sensor.asm"

.include "spi.asm"
.include "can.asm"

.include "ne2_8.asm"
.include "inet.asm"


; *
; *  return cached sensor value

can_sns_val:
	cpi	dh,	can_addr
	breq	can_sns_val_lcl
	pushy
	sety	sns_ent
	mov	r16,	dl
	swap	r16
	andi	r16,	0x0f
can_sns_val_fnd:
	ldd	r17,	Y+0x0e
	cpse	r17,	dh
	rjmp	can_sns_val_nxt
	ldd	r17,	Y+0x0f
	cpse	r17,	r16
	rjmp	can_sns_val_nxt
	andi	dl,	0x0f
	add	YL,	dl
	ld	dl,	Y
	popy
	ret
can_sns_val_nxt:
	adiw	YL,	16
	cpi	YH,	HIGH( sns_ent ) + 1
	brne	can_sns_val_fnd
	popy
	clr	dl
	ret
can_sns_val_lcl:
	mov	dh,	dl
	jmp	msr_get_sns_val



myip:
.db	IP0,	IP1,	IP2,	IP3

mymac:
.db	MAC0,	MAC1,	MAC2,	MAC3,	MAC4,	MAC5


