
; *
; *  ASM CAN MCP2515 FIELD13 PROTOCOL
; *


.include "can.def"

; *
; *  reset and setup can bus controller
	
can_init:
	rcall	spi_master
	
	sbi	CAN_PRT,	CAN_CS
	sbi	CAN_DDR,	CAN_CS

	; reset can bus controler	
	cbi	CAN_PRT,	CAN_CS
	ldi	dl,	0b11000000
	rcall	spi_trans
	sbi	CAN_PRT,	CAN_CS

	; enable pin functions
	ldi	dh,	0x0c
	ldi	dl,	0b00001100
	rcall	can_write
	ldi	dh,	0x0d
	ldi	dl,	0b00000000
	rcall	can_write
	
	cbi	CAN_PRT,	CAN_CS
	ldi	dl,	0b00000101
	rcall	spi_trans
	ldi	dl,	0x0c
	rcall	spi_trans
	ldi	dl,	0b00111100
	rcall	spi_trans
	ldi	dl,	0b00101100
	rcall	spi_trans
	sbi	CAN_PRT,	CAN_CS

	; set message filter
	setz	( can_filter_dat * 2 )
	ldi	dh,	0x60
	lpm	dl,	Z+
	rcall	can_write
	ldi	dh,	0x70
	lpm	dl,	Z+
	rcall	can_write
	ldi	dh,	0x00
	ldi	dl,	12
	rcall	can_write_block_z
	ldi	dh,	0x10
	ldi	dl,	12
	rcall	can_write_block_z
	ldi	dh,	0x20
	ldi	dl,	8
	rcall	can_write_block_z
	
	; set medium speed baud rate
	ldi	dh,	0x2a
	ldi	dl,	0b11000010
	rcall	can_write
	ldi	dh,	0x29
	ldi	dl,	0b11110011
	rcall	can_write
	ldi	dh,	0x28
	ldi	dl,	0b00000101
	rcall	can_write
	
	; set normal op mode
	ldi	dh,	0x0f
	ldi	dl,	0b00000000
	rcall	can_write
	
	; propagate station id
	setz	( can_msg_prp * 2 )
	rjmp	can_snd_z


; *
; *  read one register

can_read:
	cbi	CAN_PRT,	CAN_CS
	ldi	dl,	0b00000011
	rcall	spi_trans
	mov	dl,	dh
	rcall	spi_trans
	rcall	spi_trans
	sbi	CAN_PRT,	CAN_CS
	ret
	

; *
; *  load message from can receive buffer

can_rcv:
	rcall	can_rx_status
	sbrc	dl,	6
	rjmp	can_rcv0
	sbrc	dl,	7
	rjmp	can_rcv1
	sez
	ret
	
can_rcv0:
	ldi	dl,	0b10010000
	rjmp	can_rcv_cpy
can_rcv1:
	ldi	dl,	0b10010100
can_rcv_cpy:
	sety	can_dat
	cbi	CAN_PRT,	CAN_CS
	rcall	spi_trans
	
	ldi	dh,	0x0d
can_rcv_cpy1:
	rcall	spi_trans
	st	Y+,	dl
	dec	dh
	brne	can_rcv_cpy1
	sbi	CAN_PRT,	CAN_CS
	clz
	ret
	

; *
; *  write one register

can_write:
	push	dl
	cbi	CAN_PRT,	CAN_CS
	ldi	dl,	0b00000010
	rcall	spi_trans
	mov	dl,	dh
	rcall	spi_trans
	pop	dl
	rcall	spi_trans
	sbi	CAN_PRT,	CAN_CS
	ret


; *
; *  write register block with flash data

can_write_block_z:
	push	dl
	cbi	CAN_PRT,	CAN_CS
	ldi	dl,	0b00000010
	rcall	spi_trans
	mov	dl,	dh
	rcall	spi_trans
	pop	dh
can_write_block_z_1:
	lpm	dl,	Z+
	rcall	spi_trans
	dec	dh
	brne	can_write_block_z_1
	sbi	CAN_PRT,	CAN_CS
	ret


; *
; *  send sram message

can_snd_y:
	rcall	can_init_tx
	push	dh
	ldi	dh,	0x0d
can_snd_y1:
	ld	dl,	Y+
	rcall	spi_trans
	dec	dh
	brne	can_snd_y1
	sbi	CAN_PRT,	CAN_CS

	pop	dl	
	rjmp	can_req2snd
	

; *
; *  send static flash message

can_snd_z:
	rcall	can_init_tx
	push	dh
	ldi	dh,	0x0d
can_snd_z1:
	lpm	dl,	Z+
	rcall	spi_trans
	dec	dh
	brne	can_snd_z1
	sbi	CAN_PRT,	CAN_CS
	
	pop	dl
	rjmp	can_req2snd
	

; *
; *  select empty tx buffer

can_init_tx:
	rcall	can_status

	sbrs	dl,	2
	rjmp	can_init_tx_1
	sbrs	dl,	4
	rjmp	can_init_tx_2
	sbrs	dl,	6
	rjmp	can_init_tx_3
	
	ldi	dh,	0x30
	clr	dl
	rcall	can_write
	ldi	dh,	0x40
	clr	dl
	rcall	can_write
	ldi	dh,	0x50
	clr	dl
	rcall	can_write
	
can_init_tx_3:
	ldi	dl,	0b01000100
	ldi	dh,	0b100
	rjmp	can_init_tx_fin
can_init_tx_2:
	ldi	dl,	0b01000010
	ldi	dh,	0b010
	rjmp	can_init_tx_fin
can_init_tx_1:
	ldi	dl,	0b01000000
	ldi	dh,	0b001

can_init_tx_fin:
	cbi	CAN_PRT,	CAN_CS
	rjmp	spi_trans


; *
; *  request to send

can_req2snd:
	cbi	CAN_PRT,	CAN_CS
	andi	dl,	0b00000111
	ori	dl,	0b10000000
	rcall	spi_trans
	sbi	CAN_PRT,	CAN_CS
	ret
	

; *
; *  read can status

can_status:
	cbi	CAN_PRT,	CAN_CS
	ldi	dl,	0b10100000
	rcall	spi_trans
	rcall	spi_trans
	sbi	CAN_PRT,	CAN_CS
	ret


; *
; *  read can receive buffer status

can_rx_status:
	cbi	CAN_PRT,	CAN_CS
	ldi	dl,	0b10110000
	rcall	spi_trans
	rcall	spi_trans
	sbi	CAN_PRT,	CAN_CS
	ret


.include	"can.inc"


