
/*
	confbin
	
	option --compile <dir>
	this converts all the html, jpeg and css files
	into a binary data file html.dat
	
	option --upload <host>
	uploading to the field13 web server is also supported
	
	file format is:
	
	first 1 kb	- fat entries
	rest	- file entries
	
	fat entry	- 2 bytes address of file + 14 bytes filename
	file entry	- 2 bytes end of file address + file data
		  ( 0 end address means dynamic html data terminated by 0 )
	
*/

#include <stdio.h>
#include <stdlib.h>

#include <dirent.h>

const char * html_header = "HTTP/1.0 202 OK\r\nContent-Type: text/html\r\n\r\n";
const char * css_header = "HTTP/1.0 202 OK\r\nContent-Type: text/css\r\n\r\n";
const char * jpeg_header = "HTTP/1.0 202 OK\r\nContent-Type: image/jpeg\r\n\r\n";

struct pent {
	char * name;
	int ofs;
	struct pent * next;
};

int	cnt = 0;

unsigned char	dat[ 65536*2 ];

void	db( unsigned char c )
{
	dat[ cnt ] = c;
	cnt++;
}

void	dbalign( unsigned char c )
{
	if ( cnt & 1 != 0 )
	{
		dat[ cnt ] = c;
		cnt++;
	}
}

int	compile( char * x )
{
	printf( "compiling directory %s\n", x );
	
	DIR * D = opendir( x );
	if ( ! D )
	{
		perror( "could not read directory" );
		return -1;
	}
	
	struct pent * parr = 0;
	struct pent * pptr = parr;
	
	for ( cnt = 0; cnt < 65536*2; cnt++ )
		dat[ cnt ] = 0xff;
	
	cnt = 1024;
	
	struct dirent * d = readdir( D );
	while ( d )
	{
		int l = strlen( d->d_name );
		
		const char * cont = 0;
		
		if ( l > 5 && ( strcmp( d->d_name + l - 5, ".html" ) == 0 ) )
			cont = html_header;
		if ( l > 5 && ( strcmp( d->d_name + l - 4, ".jpg" ) == 0 ) )
			cont = jpeg_header;
		if ( l > 5 && ( strcmp( d->d_name + l - 4, ".css" ) == 0 ) )
			cont = css_header;
		if ( cont )
		{
			struct pent * p = (struct pent *) malloc( sizeof( struct pent ) );
			
			p->name	= (char *) malloc( l + 1 );
			
			strcpy( p->name, d->d_name );
			
			p->ofs	= cnt;
			
			p->next = 0;
			
			if ( ! parr )
			{
				parr = p;
				pptr = p;
			}
			else
			{
				pptr->next = p;
				pptr = p;
			}


			printf( "%s - ", pptr->name );
			
			db( 0 );
			db( 0 );
			for ( l = 0; l < strlen( cont ); l++ )
			{
				db( cont[ l ] );
			}
		
			unsigned char c;
			
			char * fname = (char *) malloc( strlen( x ) + strlen( pptr->name ) + 1 );
			strcpy( fname, x );
			strcpy( fname + strlen( x ), pptr->name );
			
			FILE * F = fopen( fname, "r" );
			if ( ! F )
			{
				perror( "could not read file" );
				return -1;
			}
			
			free( fname );
			
			if ( cont == html_header )
			{
				while ( fread( (void *) &c, 1, 1, F ) == 1 )
				{
					if ( c == (unsigned char) '$' )
					{
						unsigned char x, y;
						if ( fread( (void *) &x, 1, 1, F ) == 1 && fread( (void *) &y, 1, 1, F ) == 1 )
						{
							x = ( x >= 'A' ? x - 'A' + 0xA : x - '0' );
							y = ( y >= 'A' ? y - 'A' + 0xA : y - '0' );
							
							int z;
							z = x & 0x0f;
							z <<= 4;
							z += y & 0x0f;
							
							dbalign( ' ' );
							db( c );
							db( z );
						}
					}
					else if ( c == (unsigned char) '' )
					{
						unsigned char x[4];
						if ( fread( (void *) x, 1, 4, F ) == 4 )
						{
							x[ 0 ] = ( x[ 0 ] >= 'A' ? x[ 0 ] - 'A' + 0xA : x[ 0 ] - '0' );
							x[ 1 ] = ( x[ 1 ] >= 'A' ? x[ 1 ] - 'A' + 0xA : x[ 1 ] - '0' );
							x[ 2 ] = ( x[ 2 ] >= 'A' ? x[ 2 ] - 'A' + 0xA : x[ 2 ] - '0' );
							x[ 3 ] = ( x[ 3 ] >= 'A' ? x[ 3 ] - 'A' + 0xA : x[ 3 ] - '0' );
							
							dbalign( ' ' );
							db( c );
							db( 0 );
							int z = x[ 0 ] & 0x0f;
							z <<= 4;
							z += x[ 1 ] & 0x0f;
							db( z );
							z = x[ 2 ] & 0x0f;
							z <<= 4;
							z += x[ 3 ] & 0x0f;
							db( z );
						}
					}
					else
						db( c );
				}
				dbalign( ' ' );
				db( 0 );
				db( 0 );
			}
			else
			{
				while ( fread( (void *) &c, 1, 1, F ) == 1 )
				{
					db( c );
				}
				dbalign( 0x00 );
				dat[ pptr->ofs ] = ( cnt / 2 ) >> 8 & 0xff;
				dat[ pptr->ofs+1 ] = ( cnt / 2 ) & 0xff;
			}
		
			fclose( F );

			printf( "\tofs %02x%02x\n", ( p->ofs / 2 ) >> 8 & 0xff, ( p->ofs / 2 ) & 0xff );
		}
			
		d = readdir( D );
	}
	
	pptr = parr;
	int pcnt = 0;

	if ( ! pptr )
	{
		perror( "no pages found" );
		return -1;
	}
	
	cnt = 0;
	
	while ( pptr )
	{
		db( ( pptr->ofs / 2 ) >> 8 & 0xff );
		db( ( pptr->ofs / 2 ) & 0xff );
		
		int l = strlen( pptr->name );
		
		int c;
		for ( c = 0; c < 14; c++ )
		{
			if ( c < l )
				db( pptr->name[ c ] );
			else
				db( 0 );
		}
			
		pptr = pptr->next;	
	}
	db( 0 );
	db( 0 );
	
	if ( cnt > 1024 )
	{
		fprintf( stderr, "too many files" );
		return -1;
	}


	FILE * F = fopen( "html.dat", "w" );
	if ( ! F )
	{
		perror( "could not write file" );
		return -1;
	}
	
	if ( fwrite( (void *) dat, 1, 65536*2, F ) != 65536*2 )
	{
		perror( "failed to write file" );
		return -1;
	}
	
	fclose( F );
	
	return 0;
}

#include <arpa/inet.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>

int	upload( char * x )
{
	printf( "uploading to %s:%u\n", x, 42135 );

	struct sockaddr_in addr;
	addr.sin_family = AF_INET;
	addr.sin_port = (unsigned short) htons( 42135 );
	
	if ( inet_aton( x, &addr.sin_addr ) == 0 )
	{
		printf( "invalid address\n" );
		return -1;
	}

	int udps = socket( PF_INET, SOCK_DGRAM, 0 );
	
	if ( connect( udps, (struct sockaddr *) &addr, sizeof( addr ) ) < 0 )
	{
		perror( "connect" );
		return -1;
	}
	
	fcntl( udps, F_SETFL, O_NONBLOCK );
	//fcntl( 1, F_SETFL, O_DIRECT );
	
	FILE * F = fopen( "html.dat", "r" );
	if ( ! F )
	{
		perror( "could not open html.dat" );
		return -1;
	}
	
	if ( fread( (void *) dat, 1, 65536*2, F ) != 65536*2 )
	{
		printf( "html.dat : file size mismatch\n" );
		return -1;
	}
	
	fclose( F );
	
	unsigned int ofs = 0;
	int retry = 0;
	int i, j;
	
	unsigned char * smsg = (unsigned char *) malloc( 128+3 );
	unsigned char * rmsg = (unsigned char *) malloc( 255 );
	
	if ( ! smsg || ! rmsg )
	{
		printf( "message buffer allocation failed\n" );
		return -1;
	}
	
	printf( "\n" );
	
	while ( ofs < 65536 * 2 && retry < 3 )
	{
		smsg[ 0 ] = 0x82;
		smsg[ 1 ] = ( ofs / 2 ) >> 8 & 0xff;
		smsg[ 2 ] = ( ofs / 2 ) & 0xff;
		memcpy( &smsg[3], &dat[ofs], 128 );
		printf( "ofs [ %02x %02x ]\r", smsg[ 1 ], smsg[ 2 ] );

		send( udps, smsg, 128+3, 0 );
		
		for ( i = 0; i < 1000; i++ )
		{
			int r = read( udps, rmsg, 255 );
			if ( r == 128 + 3 )
			{
				for ( j = 0; j < 128 + 3; j++ )
				{
					if ( smsg[ j ] != rmsg[ j ] )
					{
						printf( "\nverify failed (msg ofs %d)\n", j );
						return -1;
					}
				}
				ofs += 128;
				break;
			}
			else if ( r != -1 )
			{
				printf( " ( read %d ) ", r );
			}
			usleep( 1000 );
		}
		if ( i >= 1000 )
		{
			retry++;
			printf( "\ntimeout retries left : %d\n", 3 - retry );
		}
	}
	
	if ( retry >= 3 )
	{
		printf( "\nretry count exceeded - upload failed\n" );
		return -1;
	}
	
	printf( "\n\nupload complete\n" );
	
	return 0;
}

void	usage()
{
	printf( "\nusage: confbin [--compile directory] [--upload hostaddress]\n\n" );
}

int	main( int argc, char * argv[] )
{
	char * d = 0;
	char * h = 0;
	
	int i, r;
	
	for ( i = 1; i < argc; i++ )
	{
		if ( strcmp( argv[ i ], "--compile" ) == 0 )
		{
			if ( i+1 < argc )
			{
				d = argv[ ++i ];
			}
			else
			{
				usage();
				return  -1;
			}
		}
		else if ( strcmp( argv[ i ], "--upload" ) == 0 )
		{
			if ( i + 1 < argc )
			{
				h = argv[ ++i ];
			}
			else
			{
				usage();
				return -1;
			}
		}
		else
		{
			usage();
			return -1;
		}
	}
	
	if ( !d && !h )
	{
		usage();
		return -1;
	}
	
	if ( d )
	{
		r = compile( d );
		if ( r != 0 )
		{
			return r;
		}
	}
	
	if ( h )
	{
		r = upload( h );
		if ( r != 0 )
		{
			return r;
		}
	}
	
	return 0;
}
