
; *
; *  ASM AVR FIELD13 HTTP WEB SERVER
; *


; page hits
.equ	http_count	= ( nic_dat - 2 )


; *
; *  direct write one data word to nic transmit buffer
	
.macro	http_write_word_m

	nic_dma_write_r	dh
	nic_dma_write_r dl
	
	adiw	XL,	2

	; update checksum on the fly
	add	r1,	dl
	adc	r2,	dh
	adc	r1,	r0
	adc	r2,	r0
.endmacro

http_write_word:
	http_write_word_m
	ret


; *
; *  clear access counter

http_init:
	clr	r16
	sts	http_count,	r16
	ret


; *
; *  process incoming http packets

http_in:
	; get tcp connection data
	rcall	tcp_sav_get
	brts	http_chk_sts

http_syn:
	; not found -> new connection
	sety	nic_dat
	ldd	r16,	Y+TCP_FLG
	cpi	r16,	( 1 << TCP_FLG_SYN )
	breq	http_syn_do
	rjmp	tcp_reset
http_syn_do:
	rcall	tcp_sav_new
	brtc	http_syn_drop

	; store new ack number
	rcall	tcp_sav_ack_cpy
	setx	1
	rcall	tcp_sav_ack_add

	; ack syn
	ldi	r16,	( 1 << TCP_FLG_SYN ) | ( 1 << TCP_FLG_ACK )
	;rjmp	http_snd_ack
http_snd_ack:
	clr	XL
	clr	XH
	rcall	tcp_header
	rcall	tcp_chksum

	sety	nic_dat
	std	Y+TCP_CHK+0,	dh
	std	Y+TCP_CHK+1,	dl
	rjmp	ip_out
http_syn_drop:
	ret

	
http_chk_sts:
	; connection was found -> check flags
	
	sety	nic_dat
	ldd	r16,	Y+TCP_FLG
	
	sbrc	r16,	TCP_FLG_RST
	rjmp	http_sts_rst
	
	sbrc	r16,	TCP_FLG_SYN
	rjmp	http_sts_syn
	
	sbrc	r16,	TCP_FLG_FIN
	rjmp	http_sts_fin
	
	sbrc	r16,	TCP_FLG_ACK
	rjmp	http_sts_ack

http_sts_fin:
http_sts_rst:
http_sts_syn:
	; release connection
	rjmp	tcp_sav_clr
	
http_sts_ack:
	; ack only received -> check sequence number
	rcall	tcp_sav_nsq_chk
	brne	http_old_seq_chk

	; update sequence & page pointers
	rcall	tcp_sav_seq_cpy
	
	ldd	dl,	Z+tcp_sav_dat+4
	std	Z+tcp_sav_dat+0,	dl
	ldd	dl,	Z+tcp_sav_dat+5
	std	Z+tcp_sav_dat+1,	dl

	rjmp	http_seq_match

http_old_seq_chk:
	; check last sequence number
	rcall	tcp_sav_seq_chk
	brne	http_sts_rst

http_seq_match:
	; update ack number
	rcall	tcp_dat_len
	rcall	tcp_sav_ack_cpy
	rcall	tcp_sav_ack_add
	
	; check connection state
	ldd	r16,	Z+tcp_sav_sts
	cpi	r16,	0x10
	brcc	http_continue

	; check if data present
	tst	XL
	brne	http_start
	tst	XH
	brne	http_start
	
	ret

http_start:
	; push tcp connection pointer
	pushz
	
	; evaluate optional queries
	rcall	http_query
	
	; select the right page
	rcall	http_page_select
	
	; load page end pointer
	rcall	i2c_read
	rcall	i2c_in
	mov	r3,	dh
	mov	r4,	dl
	adiw	ZL,	1
	
	; save page pointers
	popy
	std	Y+tcp_sav_dat+0,	ZH
	std	Y+tcp_sav_dat+1,	ZL
	std	Y+tcp_sav_dat+2,	r3
	std	Y+tcp_sav_dat+3,	r4

	; update connection state
	ldi	r16,		0x10
	std	Y+tcp_sav_sts,		r16

	rjmp	http_write_data	
	
http_continue:
	; load page pointers
	ldd	dh,	Z+tcp_sav_dat+0
	ldd	dl,	Z+tcp_sav_dat+1
	ldd	r3,	Z+tcp_sav_dat+2
	ldd	r4,	Z+tcp_sav_dat+3
	
	; check if page pointer at end
	cp	dh,	r3
	brne	http_do_continue
	cp	dl,	r4
	brne	http_do_continue	
	
	; send fin
	pushz
	ldi	r16,	( 1 << TCP_FLG_FIN ) | ( 1 << TCP_FLG_ACK )
	rcall	http_snd_ack
	popz
	rjmp	tcp_sav_clr

http_do_continue:
	; save tcp connection pointer
	mov	YH,	ZH
	mov	YL,	ZL
	; page pointer is Z
	mov	ZH,	dh
	mov	ZL,	dl

	; start reading data
	rcall	i2c_read
	
http_write_data:
	; set dma start address
	ldi	r16,	TCP_DAT
	rcall	nic_dma_write_start
	
	; clear counter & checksum
	clr	XL
	clr	XH
	clr	r0
	clr	r1
	clr	r2
	
	; zero end adress indicates dynamic html data
	tst	r3
	breq	http_write_dyn_page
	
	; copy binary data
http_write_page_data:
	rcall	i2c_in
	
	http_write_word_m

	adiw	ZL,	1

	; MTU reached ?
	cpi	XH,	4
	breq	http_write_page_data_jmp_done

	; page end reached ?
	cp	ZH,	r3
	brne	http_write_page_data
	cp	ZL,	r4
	brne	http_write_page_data

http_write_page_data_jmp_done:
	rjmp	http_write_page_data_done

	; write dynamic data (check for some special characters)
http_write_dyn_page:
	rcall	i2c_in

	tst	dh
	breq	http_write_dyn_page_done

	adiw	ZL,	1

	cpi	dh,	'$'
	breq	http_write_dyn_page_insert
	cpi	dh,	''
	breq	http_write_dyn_page_insert_sns

	http_write_word_m

	cpi	XH,	4
	brne	http_write_dyn_page

	rjmp	http_write_page_data_done

http_write_dyn_page_done:
	clr	ZH
	clr	ZL
	rjmp	http_write_page_data_done

http_write_dyn_page_insert:
	; insert special vars
	cpi	dl,	'C'
	breq	http_write_dyn_page_insert_count

	rcall	http_write_hexbyte
	rjmp	http_write_dyn_page

http_write_dyn_page_insert_count:
	lds	dl,	http_count
	inc	dl
	sts	http_count,	dl
	rcall	http_write_hexbyte
	rjmp	http_write_dyn_page

http_write_dyn_page_insert_sns:
	; insert can sensor values
	rcall	i2c_in
	adiw	ZL,	1

	rcall	can_sns_val
	rcall	http_write_hexbyte
	rjmp	http_write_dyn_page

http_write_page_data_done:
	; complete dma transfer
	rcall	i2c_stop
	rcall	nic_dma_write_stop

	; push checksum
	push	r1
	push	r2

	; save current page pointer
	std	Y+tcp_sav_dat+4,	ZH
	std	Y+tcp_sav_dat+5,	ZL
;	mov	dh,	ZH
;	mov	dl,	ZL

	; restore tcp sav pointer
	mov	ZH,	YH
	mov	ZL,	YL

	; update sequence number
	rcall	tcp_sav_nsq_add

	; set up tcp header
	ldi	r16,	( 1 << TCP_FLG_PSH ) | ( 1 << TCP_FLG_ACK )
;	cp	dh,	r3
;	brne	http_write_fin_skip
;	cp	dl,	r4
;	brne	http_write_fin_skip
;	ori	r16,	( 1 << TCP_FLG_FIN )
;http_write_fin_skip:
	rcall	tcp_header

	; calculate tcp header checksum
	pop	dh
	pop	dl
	rcall	tcp_chksum_pseudo
	ldi	XL,	(TCP_DAT-IP_DAT)
	clr	XH
	sety	nic_dat+IP_DAT
	rcall	inet_chksum_part
	ser	r16
	eor	dh,	r16
	eor	dl,	r16
	; store checksum
	sety	nic_dat
	std	Y+TCP_CHK+0,	dh
	std	Y+TCP_CHK+1,	dl
	
	rcall	ip_chksum
	
	; write eth + ip + tcp header
	ldi	XL,	TCP_DAT
	clr	XH
	clr	r16
	rcall	nic_dma_write_start
	rcall	nic_dma_write_buffer
	rcall	nic_dma_write_stop
	
	; start transmition
	sety	nic_dat
	ldd	XH,	Y+IP_LEN+0
	ldd	XL,	Y+IP_LEN+1
	adiw	XL,	ETH_DAT
	rjmp	nic_trx


; *
; *  check http get command

http_page_select:
	; check if GET command is present in tcp data
	rcall	tcp_dat_ptr
	setz	( http_get_string * 2 )
	rcall	http_match
	brne	http_page_select_index
	
	; start reading fat entries
	clr	ZH
	clr	ZL
	rcall	i2c_read
	
http_page_select_check_file:
	; load page pointer
	rcall	i2c_in
	mov	ZH,	dh
	mov	ZL,	dl
	
	tst	ZH
	breq	http_page_select_index_stop
	
	; file name length maximum is 14 bytes 
	ldi	r17,	7
	
	pushy
	
	; carefully name your files:
	; i, index, index.html all match GET /index.html
http_page_select_check_char:
	; read in two characters and check them
	rcall	i2c_in

	tst	dh
	breq	http_page_select_match
	
	ld	r18,	Y+
	
	cp	dh,	r18
	brne	http_page_select_skip
	
	ld	r18,	Y+
	
	tst	dl
	breq	http_page_select_match
	
	cp	dl,	r18
	brne	http_page_select_skip
	
	dec	r17
	brne	http_page_select_check_char
	
http_page_select_match:
	; return page pointer in ZH:ZL
	rcall	i2c_stop
	popy
	ret
	
http_page_select_index_stop:
	rcall	i2c_stop
http_page_select_index:
	; return default page
	ldi	ZH,	0x02
	ldi	ZL,	0x00
	ret

http_page_select_skip:
	; skip the rest of the current file name
	dec	r17
	breq	http_page_select_next
	
	rcall	i2c_in
	rjmp	http_page_select_skip
	
http_page_select_next:
	; restore tcp data pointer with GET string
	popy
	; and check next file
	rjmp	http_page_select_check_file
	
http_get_string:
.db	"GET /", 0


; *
; *  evaluate optional http query string ( index.html?query=xyz )
	
http_query:
	rcall	tcp_dat_ptr
	
	setz	( query_sensor * 2 )
	rcall	http_match
	breq	http_query_sensor
http_query_done:
	ret
	
http_query_sensor:
	; convert can address from ascii hex to binary
	ld	dh,	Y+
	ld	dl,	Y+

	subi	dh,	'0'
	cpi	dh,	0x0a
	brcs	http_query_sensor_1
	subi	dh,	7
http_query_sensor_1:
	subi	dl,	'0'
	cpi	dl,	0x0a
	brcs	http_query_sensor_2
	subi	dl,	7
http_query_sensor_2:
	swap	dh
	or	dh,	dl

	; setup can sensor query command message
	setz	can_dat
	st	Z+,	dh
	ldi	dh,	can_addr_ext
	st	Z+,	dh
	
	; convert sensor number from ascii hex to bin
	ld	dh,	Y+
	subi	dh,	'0'
	cpi	dh,	0x0a
	brcs	http_query_sensor_3
	subi	dh,	7
http_query_sensor_3:
	ori	dh,	cmd_sns_0
	
	; complete message
	st	Z+,	dh
	clr	dl
	st	Z+,	dl
	ldi	dl,	0b01000000
	st	Z+,	dl
	
	; and send it down the can bus
	sety	can_dat
	rjmp	can_snd_y


; *
; *  compare HTTP command line with flash string

http_match:
	pushy
	pushz
http_match_1:
	lpm	dh,	Z+
	tst	dh
	breq	http_match_fin
	
http_match_2:
	ld	dl,	Y+
	cp	dl,	dh
	breq	http_match_1
	cpi	dl,	'\r'
	breq	http_match_no
	popz
	pushz
	rjmp	http_match_1

http_match_no:
	clz
	popy
	popy		; restore y
	ret
http_match_fin:
	pop	dl	; restore nothing
	pop	dh
	pop	dl
	pop	dh
	ret


; *
; *  write dl in hex form

http_write_hexbyte:
	mov	dh,	dl
	andi	dh,	0xf0
	swap	dh
	cpi	dh,	0x0a
	ldi	r16,	'0'
	brcs	http_write_hexbyte_h1
	ldi	r16,	('A' - 0x0a)
http_write_hexbyte_h1:
	add	dh,	r16
	andi	dl,	0x0f
	ldi	r16,	'0'
	cpi	dl,	0x0a
	brcs	http_write_hexbyte2
	ldi	r16,	('A' - 0x0a)
http_write_hexbyte2:
	add	dl,	r16
	rcall	http_write_word
	ret


; *
; *  query commands

;query_display:
;.db	"display=", 0

query_sensor:
.db	"sensor=", 0
