
; *
; *  dual parallel i2c software interface
; *


.equ	I2C_PRT	=	PORTB
.equ	I2C_PIN	=	PINB
.equ	I2C_DDR	=	DDRB

.equ	I2C_SCL	=	PB2	; first i2c channel
.equ	I2C_SDA	=	PB3
.equ	I2C_TDA	=	PB1	; second i2c channel
;.equ	I2C_TCL	=	PB0
.equ	I2C_WPRT	=	PB0	; write protect pin

; read/write commands (bus address is 0)
.equ	I2C_RD	=	0b10100001
.equ	I2C_WR	=	0b10100000


; *
; *  macros

.macro	SDA_HI
	sbi	I2C_PRT,	I2C_SDA
.endmacro
.macro	TDA_HI
	sbi	I2C_PRT,	I2C_TDA
.endmacro

.macro	SDA_LO
	cbi	I2C_PRT,	I2C_SDA
.endmacro
.macro	TDA_LO
	cbi	I2C_PRT,	I2C_TDA
.endmacro

.macro	SCL_HI
	sbi	I2C_PRT,	I2C_SCL
;	sbi	I2C_PRT,	I2C_TCL
.endmacro

.macro	SCL_LO
	cbi	I2C_PRT,	I2C_SCL
;	cbi	I2C_PRT,	I2C_TCL
.endmacro

.macro	I2C_DLY
	; clock speed 8 mhz - eeprom 2 * atmel 24c512 (1Mbit/s)
	nop
	nop
	nop
	nop
.endmacro

.macro	I2C_CLK
	; strobe i2c clock
	SCL_HI
	I2C_DLY
	SCL_LO
.endmacro


; *
; *  i2c port setup

i2c_init:
	sbi	I2C_DDR,	I2C_SCL
	sbi	I2C_DDR,	I2C_SDA
	sbi	I2C_DDR,	I2C_TDA
	cbi	I2C_DDR,	I2C_WPRT
	cbi	I2C_PRT,	I2C_WPRT
	
	SDA_HI
	TDA_HI
	SCL_HI
	ret


; *
; *  generate start condition
	
i2c_start:
	SDA_HI
	TDA_HI
	SCL_HI
	SDA_LO
	TDA_LO
	SCL_LO
	ret


; *
; *  generate stop condition

i2c_stop:
	SDA_LO
	TDA_LO
	SCL_HI
	SDA_HI
	TDA_HI
	ret


; *
; *  write two bytes at once (raid0)

i2c_out:
	push	r16
	ldi	r16,	8
i2c_out_bit:
	SDA_HI
	rol	dl
	brcs	i2c_out_bit2
	SDA_LO
i2c_out_bit2:
	TDA_HI
	rol	dh
	brcs	i2c_out_clk
	TDA_LO
i2c_out_clk:
	I2C_CLK
	dec	r16
	brne	i2c_out_bit
	
	cbi	I2C_DDR,	I2C_SDA
	cbi	I2C_DDR,	I2C_TDA
	SDA_HI
	TDA_HI
	SCL_HI
	I2C_DLY
	set
	sbic	I2C_PIN,	I2C_SDA
	clt
	sbic	I2C_PIN,	I2C_TDA
	clt
	SCL_LO
	sbi	I2C_DDR,	I2C_SDA
	sbi	I2C_DDR,	I2C_TDA
	
	pop	r16
	ret


; *
; *  read two bytes at once

i2c_in:
	push	r16
	ldi	r16,	8
	cbi	I2C_DDR,	I2C_SDA
	cbi	I2C_DDR,	I2C_TDA
i2c_in_bit:
	SCL_HI
	clc
	sbic	I2C_PIN,	I2C_SDA
	sec
	rol	dl
	clc
	sbic	I2C_PIN,	I2C_TDA
	sec
	rol	dh
	SCL_LO
	dec	r16
	brne	i2c_in_bit
	
	sbi	I2C_DDR,	I2C_SDA
	sbi	I2C_DDR,	I2C_TDA
	SDA_LO
	TDA_LO
	I2C_CLK
	SDA_HI
	TDA_HI
	
	pop	r16
	ret


; *
; *  i2c eeprom start random read at ZH:ZL

i2c_read:	
	rcall	i2c_start
	ldi	dl,	I2C_WR
	mov	dh,	dl
	rcall	i2c_out
	mov	dl,	ZH
	mov	dh,	dl
	rcall	i2c_out
	mov	dl,	ZL
	mov	dh,	dl
	rcall	i2c_out
	rcall	i2c_start
	ldi	dl,	I2C_RD
	mov	dh,	dl
	rjmp	i2c_out


; *
; *  i2c eeprom start random write at ZH:ZL

i2c_write:
	rcall	i2c_start
	ldi	dl,	I2C_WR
	mov	dh,	dl
	rcall	i2c_out
	mov	dl,	ZH
	mov	dh,	dl
	rcall	i2c_out
	mov	dl,	ZL
	mov	dh,	dl
	rjmp	i2c_out

