
; ip header
.equ	IP_VRS	=	ETH_DAT + 0x00
.equ	IP_TOS	=	ETH_DAT + 0x01
.equ	IP_LEN	=	ETH_DAT + 0x02
.equ	IP_ID	=	ETH_DAT + 0x04
.equ	IP_FRG	=	ETH_DAT + 0x06
.equ	IP_TTL	=	ETH_DAT + 0x08
.equ	IP_PRT	=	ETH_DAT + 0x09
.equ	IP_CHK	=	ETH_DAT + 0x0a
.equ	IP_SRC	=	ETH_DAT + 0x0c 
.equ	IP_DST	=	ETH_DAT + 0x10
.equ	IP_DAT	=	ETH_DAT + 0x14

; ip sub protocol types
.equ	IP_PRT_ICMP	=	0x01
.equ	IP_PRT_TCP	=	0x06
.equ	IP_PRT_UDP	=	0x11


; *
; *  process ip packet

ip_in:
	; check ip version and header length
	lds	dl,	nic_dat+IP_VRS
	cpi	dl,	0x45
	brne	ip_drop
	
	; verify header checksum
	sety	nic_dat
	clr	XH
	ldi	XL,	20
	sety	( nic_dat + ETH_DAT )
	rcall	inet_chksum
	sety	nic_dat
	tst	dl
	brne	ip_drop
	tst	dh
	brne	ip_drop

ip_chk_icmp:	
	; check for icmp protocol
	sety	nic_dat
	ldd	r16,	Y+IP_PRT
	cpi	r16,	IP_PRT_ICMP
	brne	ip_chk_ip_addr
	rjmp	icmp_in
	
ip_chk_ip_addr:
	; check ip target address
	sety	( nic_dat+IP_DST )
	setz	( myip * 2 )
	setx	4
	rcall	proto_match
	brne	ip_drop
	
	; check sub protocol type
	sety	nic_dat
	ldd	r16,	Y+IP_PRT
ip_chk_tcp:
	cpi	r16,	IP_PRT_TCP
	brne	ip_chk_udp
	rjmp	tcp_in
ip_chk_udp:
	cpi	r16,	IP_PRT_UDP
	brne	ip_drop_proto
	rjmp	udp_in
ip_drop:
	ret
ip_drop_proto:
	rjmp	icmp_dest_proto


; *
; *  send ip packet

ip_out:
	rcall	ip_chksum
	sety	nic_dat
	ldd	XH,	Y+IP_LEN+0
	ldd	XL,	Y+IP_LEN+1
	adiw	XL,	ETH_DAT
	rjmp	nic_snd

	
; *
; *  generate ip header

ip_header:
	; target mac <- source mac <- my mac
	sety	nic_dat
	setz	( mymac * 2 )
	ldi	dh,	6
ip_head_set_mac:
	ldd	r0,	Y+ETH_SRC
	std	Y+ETH_DST,	r0
	lpm
	std	Y+ETH_SRC,	r0
	adiw	YL,	1
	adiw	ZL,	1
	dec	dh
	brne	ip_head_set_mac
	
	; destination ip <- source ip <- my ip
	sety	nic_dat
	setz	( myip * 2 )
	ldi	dh,	4
ip_head_set_ip:
	ldd	dl,	Y+IP_SRC
	std	Y+IP_DST,	dl
	lpm
	std	Y+IP_SRC,	r0
	adiw	YL,	1
	adiw	ZL,	1
	dec	dh
	brne	ip_head_set_ip
	ret
	
ip_chksum:
	sety	nic_dat
	; calculate header checksum
	clr	XH
	std	Y+IP_CHK,	XH
	std	Y+IP_CHK+1,	XH
	ldi	XL,	20
	adiw	YL,	ETH_DAT
	rcall	inet_chksum
	
	; store checksum
	sety	nic_dat
	std	Y+IP_CHK+0,	dh
	std	Y+IP_CHK+1,	dl
	
	ret

