
; *
; *  ASM AVR ISA ETHER NE2000
; *

;.equ	MAC0	=	0x00	; nic's mac adress
;.equ	MAC1	=	0x01
;.equ	MAC2	=	0x23
;.equ	MAC3	=	0x45
;.equ	MAC4	=	0x67
;.equ	MAC5	=	0x89

;.equ	nic_dat	=	0x100	; 252 bytes rx/tx buffer
.equ	nic_dat_end	=	(nic_dat+0x100)

.equ	nic_tx_page	=	(nic_dat-1)	; empty transmit page pointer


; multi purpose data registers
;.def	dl	=	r21
;.def	dh	=	r22


; port definitions
.equ	NIC_ADDR_PORT	=	PORTA
.equ	NIC_ADDR_DDR	=	DDRA

.equ	NIC_DATA_DDR	=	DDRC
.equ	NIC_DATA_PORT	=	PORTC
.equ	NIC_DATA_PIN	=	PINC

.equ	NIC_CTRL_PORT	=	PORTD
.equ	NIC_CTRL_DDR	=	DDRD
.equ	NIC_CTRL_PIN	=	PIND

.equ	NIC_INT	=	PD3
.equ	NIC_RST	=	PD4
.equ	NIC_DIS	=	PD5	; nic power supply control pin
.equ	NIC_WEN	=	PD6
.equ	NIC_REN	=	PD7


; standard ne2000 registers
.equ	NIC_CR	=	0x00
.equ	NIC_RDMA	=	0x10
.equ	NIC_RSTR	=	0x1f
; register page 0
.equ	NIC_PSTART	=	0x01
.equ	NIC_PSTOP	=	0x02
.equ	NIC_BNRY	=	0x03
.equ	NIC_TPSR	=	0x04
.equ	NIC_TBCR0	=	0x05
.equ	NIC_TBCR1	=	0x06
.equ	NIC_ISR	=	0x07
.equ	NIC_RSAR0	=	0x08
.equ	NIC_RSAR1	=	0x09
.equ	NIC_RBCR0	=	0x0a
.equ	NIC_RBCR1	=	0x0b
.equ	NIC_RCR	=	0x0c
.equ	NIC_TCR	=	0x0d
.equ	NIC_DCR	=	0x0e
.equ	NIC_IMR	=	0x0f
.equ	NIC_TLY0	=	0x05
.equ	NIC_TLY1	=	0x0d
.equ	NIC_TLY2	=	0x0e
.equ	NIC_TLY3	=	0x0f
; register page 1
.equ	NIC_PAR0	=	0x01
.equ	NIC_PAR1	=	0x02
.equ	NIC_PAR2	=	0x03
.equ	NIC_PAR3	=	0x04
.equ	NIC_PAR4	=	0x05
.equ	NIC_PAR5	=	0x06
.equ	NIC_CURR	=	0x07
.equ	NIC_MAR0	=	0x08
.equ	NIC_MAR1	=	0x09
.equ	NIC_MAR2	=	0x0a
.equ	NIC_MAR3	=	0x0b
.equ	NIC_MAR4	=	0x0c
.equ	NIC_MAR5	=	0x0d
.equ	NIC_MAR6	=	0x0e
.equ	NIC_MAR7	=	0x0f

; receive ring buffer
.equ	NIC_RXSTART	=	0x50
.equ	NIC_RXSTOP	=	0x60	; only 8kb supported in 8 bit mode

; two transmit buffers with 8 pages each
.equ	NIC_TXPAGE	=	0x40

; interrupt flags
.equ	PRX	=	0
.equ	PTX	=	1
.equ	OVFL	=	4
.equ	CNTR	=	5
.equ	RDC	=	6

; control flags
.equ	STP	=	0
.equ	STA	=	1
.equ	TXP	=	2
.equ	RD0	=	3
.equ	RD1	=	4
.equ	RD2	=	5
.equ	PS0	=	6
.equ	PS1	=	7


.equ	NIC_SEQ_READ	=	0xfe
.equ	NIC_SEQ_END	=	0xff


; *
; * some macros

.macro	nic_read
	ldi	dh,	@0
	rcall	_nic_read
.endmacro

.macro	nic_write
	ldi	dh,	@0
	ldi	dl,	@1
	rcall	_nic_write
.endmacro

.macro	nic_write_r
	ldi	dh,	@0
	mov	dl,	@1
	rcall	_nic_write
.endmacro

.macro	nic_read_r
	ldi	dh,	@0
	rcall	_nic_read
	mov	@1,	dl
.endmacro

.macro	nic_delay
	; clock speed 8 mhz
	nop
.endmacro

.macro	nic_dma_write_r
	out	NIC_DATA_PORT,	@0
	cbi	NIC_CTRL_PORT,	NIC_WEN
	nic_delay
	sbi	NIC_CTRL_PORT,	NIC_WEN
.endmacro

.macro	_nic_dma_read_byte
	cbi	NIC_CTRL_PORT,	NIC_REN
	nic_delay
	in	dl,	NIC_DATA_PIN
	sbi	NIC_CTRL_PORT,	NIC_REN
.endmacro

nic_dma_read_byte:
	_nic_dma_read_byte
	ret


; *
; * write register routine

_nic_write:
	out	NIC_ADDR_PORT,	dh
	ser	dh
	out	NIC_DATA_DDR,	dh
	out	NIC_DATA_PORT,	dl
	cbi	NIC_CTRL_PORT,	NIC_WEN
	nic_delay
	sbi	NIC_CTRL_PORT,	NIC_WEN
	ret


; *
; * read register routine

_nic_read:
	out	NIC_ADDR_PORT,	dh
	clr	dh
	out	NIC_DATA_DDR,	dh
	cbi	NIC_CTRL_PORT,	NIC_REN
	nic_delay
	in	dl,	NIC_DATA_PIN
	sbi	NIC_CTRL_PORT,	NIC_REN
	ret


; *
; *  start nic routine

nic_start:
	; reset read/write enable
	sbi	NIC_CTRL_PORT,	NIC_REN
	sbi	NIC_CTRL_PORT,	NIC_WEN

	; nic power on
	cbi	NIC_CTRL_PORT,	NIC_DIS

	; nic reset high
	sbi	NIC_CTRL_PORT,	NIC_RST

	; switch nic address/control pins to output
	in	dl,	NIC_ADDR_DDR
	ori	dl,	0x1f
	out	NIC_ADDR_DDR,	dl

	in	dl,	NIC_CTRL_DDR
	sbr	dl,	( 1 << NIC_REN ) | ( 1 << NIC_WEN ) | ( 1 << NIC_DIS ) | ( 1 << NIC_RST )
	cbr	dl,	( 1 << NIC_INT )
	out	NIC_CTRL_DDR,	dl

	; pull up for interrupt pin
	sbi	NIC_CTRL_PORT,	NIC_INT

	; wait for power good
	clr	dl
nic_init1:
	rcall	nic_wait
	dec	dl
	brne	nic_init1

	; reset low
	cbi	NIC_CTRL_PORT,	NIC_RST

	; wait for hardware reset to complete
nic_init2:
	rcall	nic_wait
	dec	dl
	brne	nic_init2

	; read + write nic reset register
	nic_read	NIC_RSTR
	rcall	_nic_write

	; wait for reset to complete
nic_init3:
	nic_read	NIC_ISR
	sbrs	dl,	7
	rjmp	nic_init3

	; set page for next transmition
	ldi	dl,	NIC_TXPAGE
	sts	nic_tx_page,	dl

	; write init sequence
	ldi	ZH,	HIGH( nic_seq_init_8390 * 2 )
	ldi	ZL,	LOW( nic_seq_init_8390 * 2 )
nic_write_seq:
	lpm	dh,	Z+
	lpm	dl,	Z+

	cpi	dh,	NIC_SEQ_END
	breq	nic_write_seq_fin
	cpi	dh,	NIC_SEQ_READ
	breq	nic_write_seq_read

	rcall	_nic_write

	rjmp	nic_write_seq
nic_write_seq_read:
	mov	dh,	dl
	rcall	_nic_read
	rjmp	nic_write_seq
nic_write_seq_fin:
	ret


; *
; *  receive package routine (read only first page)

nic_rcv:
	; check if packet available
	nic_write	NIC_CR,	( 1 << STA ) | ( 1 << RD2 )
	nic_read	NIC_ISR
	sbrc	dl,	OVFL
	rjmp	nic_rcv_err
	sbrc	dl,	CNTR
	rjmp	nic_rcv_err
	sbrc	dl,	PRX
	rjmp	nic_rcv_ok
nic_rcv_nope:
	sez
	ret
nic_rcv_err:
	rcall	nic_start
	rjmp	nic_rcv
nic_rcv_ok:
	push	r16
	push	r17

	; read bounndary page
	nic_write	NIC_CR,	( 1 << STA ) | ( 1 << RD2 )
	nic_read	NIC_BNRY
	inc	dl
	cpi	dl,	NIC_RXSTOP
	brne	nic_rcv_chk
	ldi	dl,	NIC_RXSTART
nic_rcv_chk:
	mov	r16,	dl

	; start dma transfer
	nic_write	NIC_RBCR0,	0
	nic_write	NIC_RBCR1,	1
	nic_write	NIC_RSAR0,	0
	nic_write_r	NIC_RSAR1,	r16
	nic_write	NIC_CR,	( 1 << STA ) | ( 1 << RD0 )

	; select dma register
	clr	r16
	out	NIC_DATA_DDR,	r16
	ldi	r16,	NIC_RDMA
	out	NIC_ADDR_PORT,	r16

	; read header
	rcall	nic_dma_read_byte
	rcall	nic_dma_read_byte
	mov	r17,	dl
	rcall	nic_dma_read_byte
	rcall	nic_dma_read_byte

	; read first packet page into buffer
	ldi	YH,	HIGH( nic_dat )
	ldi	YL,	LOW( nic_dat )
	ldi	r16,	252
nic_rcv_dma:
	_nic_dma_read_byte
	st	Y+,	dl
	dec	r16
	brne	nic_rcv_dma

	; clear reception flag if rx ring buffer empty
	nic_write	NIC_CR,	( 1 << STA ) | ( 1 << RD2 ) | ( 1 << PS0 )
	nic_read	NIC_CURR
	mov	r16,	dl
	nic_write	NIC_CR,	( 1 << STA ) | ( 1 << RD2 )
	cp	r16,	r17
	brne	nic_rcv_set_bnry
	nic_write	NIC_ISR,	( 1 << PRX )

nic_rcv_set_bnry:
	; update boundary page
	dec	r17
	cpi	r17,	NIC_RXSTART-1
	brne	nic_rcv_set_bnry_noc
	ldi	r17,	NIC_RXSTOP-1
nic_rcv_set_bnry_noc:
	nic_write_r	NIC_BNRY,	r17

	pop	r17
	pop	r16

	; clear zero flag
	clz
	ret


; *
; *  start dma write transfer at tx_buffer + r16

nic_dma_write_start:
	nic_write	NIC_CR,	( 1 << STA ) | ( 1 << RD2 )
	nic_write_r	NIC_RSAR0,	r16
	lds	r16,	nic_tx_page
	nic_write_r	NIC_RSAR1,	r16
	nic_write	NIC_RBCR0,	0
	nic_write	NIC_RBCR1,	0x08
	nic_write	NIC_CR,	( 1 << STA ) | ( 1 << RD1 )
	ser	r16
	out	NIC_DATA_DDR,	r16
	ldi	r16,	NIC_RDMA
	out	NIC_ADDR_PORT,	r16
	ret


; *
; *  complete dma transfer

nic_dma_write_stop:
	nic_write	NIC_CR,	( 1 << STA ) | ( 1 << RD2 )
	ret


; *
; *  dma write sram buffer

nic_dma_write_buffer:
	ld	dl,	Y+
	nic_dma_write_r	dl

	dec	XL
	brne	nic_dma_write_buffer
	subi	XH,	1
	brcc	nic_dma_write_buffer

	ret


; *
; *  send package routine

nic_snd:
	; test for minimum size
	tst	XH
	brne	nic_snd1
	cpi	XL,	60
	brcc	nic_snd1
	ldi	XL,	60
nic_snd1:
	; test for maximum size
	cpi	XH,	8
	brcs	nic_snd2
	ret
nic_snd2:

	pushx

	clr	r16

	rcall	nic_dma_write_start
	rcall	nic_dma_write_buffer
	rcall	nic_dma_write_stop

	popx

	;rjmp	nic_trx


; *
; *  start transmition

nic_trx:
	nic_write	NIC_CR,	( 1 << STA ) | ( 1 << RD2 )

	; wait until transmitter ready
nic_trx2:
	nic_read	NIC_CR
	sbrc	dl,	TXP
	rjmp	nic_trx2

	; start transmition
	ldi	dh,	NIC_TPSR
	lds	dl,	nic_tx_page
	rcall	_nic_write
	nic_write_r	NIC_TBCR0,	XL
	nic_write_r	NIC_TBCR1,	XH
	nic_write	NIC_CR,	( 1 << STA ) | ( 1 << TXP ) | ( 1 << RD2 )

	; swap tx buffer start page
	lds	dl,	nic_tx_page
	ldi	dh,	0x08
	eor	dl,	dh
	sts	nic_tx_page,	dl

	ret


; *
; *  stop nic

nic_stop:
	; wait for transmitter ready
	nic_read	NIC_CR
	sbrc	dl,	TXP
	rjmp	nic_stop

	; nic power off
	sbi	NIC_CTRL_PORT,	NIC_DIS

	; switch ports to ground
	clr	dl
	out	NIC_DATA_PORT,	dl

	in	dl,	NIC_ADDR_PORT
	andi	dl,	0xe0
	out	NIC_ADDR_PORT,	dl

	cbi	NIC_CTRL_PORT,	NIC_INT
	cbi	NIC_CTRL_PORT,	NIC_RST
	cbi	NIC_CTRL_PORT,	NIC_REN
	cbi	NIC_CTRL_PORT,	NIC_WEN

	ret


; *
; *  stupid delay

nic_wait:
	push	r16
	clr	r16

nic_wait_1:
	nic_delay
	nic_delay
	dec	r16
	brne	nic_wait_1

	pop	r16
	ret


; setup sequence for ns8390 / ne2000

nic_seq_init_8390:
.db	NIC_CR,	( 1 << STP ) | ( 1 << RD2 )	; stop nic / reg page 0
.db	NIC_PSTART,	NIC_RXSTART		; receive buffer ring
.db	NIC_PSTOP,	NIC_RXSTOP
.db	NIC_BNRY,	NIC_RXSTART
.db	NIC_TPSR,	NIC_TXPAGE		; tx start page
.db	NIC_TBCR0,	0
.db	NIC_TBCR1,	0
.db	NIC_ISR,	0xff		; reset interrupt status flags
.db	NIC_RSAR0,	0		; clear dma address
.db	NIC_RSAR1,	0
.db	NIC_RBCR0,	0x00		; clear dma counters
.db	NIC_RBCR1,	0x00
.db	NIC_RCR,	0b00000100		; accept broadcasts
.db	NIC_TCR,	0b00000000		; normal operation
.db	NIC_DCR,	0b00111000		; byte transfer, fifo threshold 8
.db	NIC_IMR,	0		; no interrupts
.db	NIC_CR,	( 1 << STP ) | ( 1 << RD2 ) | ( 1 << PS0 ) ; reg page 1
.db	NIC_PAR0,	MAC0		; set mac address
.db	NIC_PAR1,	MAC1
.db	NIC_PAR2,	MAC2
.db	NIC_PAR3,	MAC3
.db	NIC_PAR4,	MAC4
.db	NIC_PAR5,	MAC5
.db	NIC_CURR,	NIC_RXSTART+1		; set current page pointer
.db	NIC_MAR0,	0xff		; clear multicast filter
.db	NIC_MAR1,	0xff
.db	NIC_MAR2,	0xff
.db	NIC_MAR3,	0xff
.db	NIC_MAR4,	0xff
.db	NIC_MAR5,	0xff
.db	NIC_MAR6,	0xff
.db	NIC_MAR7,	0xff
.db	NIC_CR,	( 1 << STA ) | ( 1 << RD2 )	; start nic & back to page 0
.db	NIC_SEQ_READ,	NIC_TLY0		; read ( reset ) tally counters
.db	NIC_SEQ_READ,	NIC_TLY1
.db	NIC_SEQ_READ,	NIC_TLY2
.db	NIC_SEQ_READ,	NIC_TLY3
.db	NIC_SEQ_END,	0		; finish

