
#include "can.h"

#include <qheader.h>
#include <qlistview.h>
#include <qpushbutton.h>
#include <qvgroupbox.h>


#define	cmd_prp		0x01
#define	cmd_sns		0x70
#define	cmd_trg		0x80
#define cmd_trg_set	0x90

#define sns_flg_aln	0x01
#define sns_flg_alx	0x02

#define max_messages	100

#define IDCOL 2
#define UNITCOL 3
#define OFSCOL 4

#include <qsplitter.h>


CAN::CAN( QWidget * parent )
	: QHGroupBox( parent ), msgcnt( 0 )
{
	setTitle( "CAN Bus" );
	setAlignment( Qt::AlignCenter );
	
	//QSplitter * split = new QSplitter( Qt::Horizontal, this );
	
	QGroupBox * group = new QVGroupBox( this );
	group->setTitle( "Messages" );
	group->setAlignment( Qt::AlignCenter );
	
	messages = new QListView( group );
	messages->addColumn( "t", 50 );
	messages->addColumn( "station id" );
	messages->addColumn( "cmd" );
	messages->addColumn( "par" );
	messages->addColumn( "rtr" );
	messages->addColumn( "data" );
	messages->addColumn( "ide" );
	
	messages->setSorting( 0, false );
	messages->setAllColumnsShowFocus( true );
	messages->setShowSortIndicator( true );
	messages->setShowToolTips( true );
	messages->setSelectionMode( QListView::NoSelection );
	
	QPushButton * bt = new QPushButton( "clear", group );
	connect( bt, SIGNAL( clicked() ), this, SLOT( clearMessages() ) );
	
	group = new QVGroupBox( this );
	group->setTitle( "Nodes" );
	group->setAlignment( Qt::AlignCenter );
	
	nodes = new QListView( group );
	//nodes->header()->setHidden( true );
	nodes->addColumn( "name" );
	nodes->addColumn( "value" );
	nodes->addColumn( "id" );
	
	nodes->setSorting( 0, true );
	nodes->setRootIsDecorated( true );
	nodes->setAllColumnsShowFocus( true );
	nodes->setShowSortIndicator( true );
	nodes->setShowToolTips( true );
	nodes->setSelectionMode( QListView::NoSelection );
	
	connect( nodes, SIGNAL( rightButtonClicked( QListViewItem *, const QPoint &, int ) ),
		 this,	SLOT( activate( QListViewItem * ) ) );
	
	bt = new QPushButton( "populate", group );
	connect( bt, SIGNAL( clicked() ), this, SLOT( listStations() ) );
	
	timer.start( 140 );
	connect( &timer, SIGNAL( timeout() ), this, SLOT( sndtimer() ) );
}

void CAN::listStations()
{
	nodes->clear();
	msg	m;
	m.id	= 0xaa;
	m.ide	= true;
	m.cmd	= cmd_prp;
	m.par	= 0x00;
	m.rtr	= true;
	m.len	= 0;
	snd( m );
}

void	CAN::clearMessages()
{
	messages->clear();
	msgcnt = 0;
	nodes->clear();
}

void CAN::activate( QListViewItem * item )
{
	if ( ! item )
		return;
	
	QListViewItem * ci = item->parent();
	if ( ! ci )
		return;
	
	QListViewItem * ni = ci->parent();
	if ( ! ni || ni->parent() )
		return;
	
	if ( ci->text( 0 ) == "sns" )
	{
		msg m;
		m.id = ni->text( IDCOL ).toInt( 0, 16 );
		m.ide = true;
		m.cmd = cmd_sns + item->text( IDCOL ).toInt( 0, 16 );
		m.par = 0;
		m.rtr = true;
		m.len = 0;
		snd( m );
	}
	else if ( ci->text( 0 ) == "trg" )
	{
		msg m;
		m.id = ni->text( IDCOL ).toInt( 0, 16 );
		m.ide = true;
		m.cmd = cmd_trg_set + item->text( IDCOL ).toInt( 0, 16 );
		m.par = ( item->text( 1 ) == "on" ? 0 : 1 );
		m.rtr = false;
		m.len = 0;
		snd( m );
	}
}

void CAN::rcv( msg m )
{
	QString t = QString( "%1" ).arg( ++msgcnt, 0, 36 );
	while ( t.length() < 10 )	t.prepend( "0" );
	QString idd = QString( "%1" ).arg( m.id, 0, 16 );
	QString cmd = QString( "%1" ).arg( m.cmd, 0, 16 );
	QString par = QString( "%1" ).arg( m.par, 0, 16 );
	QString req = ( m.rtr ? "yes" : "no" );
	QString dat;
	for ( int i = 0; i < m.len; i++ )
		dat += QString( "%1 " ).arg( m.dat[ i ], 2, 16 );
	QString ide = ( m.ide ? "yes" : "no" );
	
	QString name;
	if ( m.id == 0xaa )
	{
		name = "broadcast";
	}
	else
	{
	QListViewItem * item = nodes->firstChild();
	while ( item && item->text( IDCOL ) != idd )
		item = item->nextSibling();

	if ( item )
		name	= item->text( 0 );
	
	switch ( m.cmd )
	{
	case cmd_prp:
		{
		name = "";
		for ( int i = 0; i < m.len; i++ )
			name += QChar( m.dat[ i ] );
		if ( !item )
		{
			item = new QListViewItem( nodes, name );
			item->setText( IDCOL, idd );
			//item->setOpen( true );
		
			if ( m.par & 0x0f )
			{
				QListViewItem * sitem = new QListViewItem( item, "sns" );
				sitem->setOpen( true );
				for ( int i = 0; i < ( m.par & 0xf ); i++ )
				{
					QListViewItem * it = new QListViewItem( sitem );
					it->setText( IDCOL, QString( "%1" ).arg( i, 0, 16 ) );
					msg n;
					n.id	= m.id;
					n.cmd	= cmd_sns + i + 8;
					n.par	= 0;
					n.rtr	= true;
					n.len	= 0;
					snd( n );
					n.cmd	= cmd_sns + i;
					snd( n );
				}
				sitem->sortChildItems( IDCOL, true );
			}
			if ( m.par & 0xf0 )
			{
				QListViewItem * titem = new QListViewItem( item, "trg" );
				titem->setOpen( true );
				for ( int i = 0; i < ( ( m.par & 0xf0 ) >> 4 ); i++ )
				{
					QListViewItem * it = new QListViewItem( titem );
					it->setText( IDCOL, QString( "%1" ).arg( i, 0, 16 ) );
					msg n;
					n.id	= m.id;
					n.cmd	= cmd_trg + i;
					n.par	= 0;
					n.rtr	= true;
					n.len	= 0;
					snd( n );
				}
				titem->sortChildItems( IDCOL, true );
			}
		}}
		break;
	case ( cmd_sns + 0 ):
	case ( cmd_sns + 1 ):
	case ( cmd_sns + 2 ):
	case ( cmd_sns + 3 ):
	case ( cmd_sns + 4 ):
	case ( cmd_sns + 5 ):
	case ( cmd_sns + 6 ):
	case ( cmd_sns + 7 ):
		if ( item )
		{
			QListViewItem * sitem = item->firstChild();
			while ( sitem && sitem->text( 0 ) != "sns" )
				sitem = sitem->nextSibling();
			if ( sitem )
				sitem = sitem->firstChild();
			
			while ( sitem && sitem->text( IDCOL ) != QString( "%1" ).arg( m.cmd - cmd_sns, 0, 16 ) )
				sitem = sitem->nextSibling();

			if ( sitem )
			{
			
				QString sname;
				for ( int i = 4; i < 8; i++ )
					sname += QChar( m.dat[ i ] );
				
				sitem->setText( 0, sname );
				int unt = sitem->text( UNITCOL ).toInt();
				int ofs = (signed char) sitem->text( OFSCOL ).toInt();
				int val = m.dat[ 0 ] + ofs;
				switch ( unt )
				{
				case 0x01:
					sitem->setText( 1, QString( "%1 V" ).arg( val * 16.0 / 100.0, 0, 'f', 1 ) );
					break;
				case 0x02:
					sitem->setText( 1, QString( "%1 A" ).arg( val / 100.0, 0, 'f', 2 ) );
					break;
				case 0x05:
					sitem->setText( 1, QString( "%1 A" ).arg( val / 10.0, 0, 'f', 2 ) );
					break;
				case 0x03:
					sitem->setText( 1, QString( "%1 C" ).arg( val ) );
					break;
				case 0x04:
					sitem->setText( 1, QString( "%1 m" ).arg( val / 16.0, 0, 'f', 1 ) );
					break;
				default:
					sitem->setText( 1, QString( "%1" ).arg( m.dat[ 0 ], 0, 16 ) );
					break;
				}
				
				QListViewItem * i = sitem->firstChild();
				while ( i && i->text( 0 ) != "alarm" )
					i = i->nextSibling();
				if ( ! i )
					i = new QListViewItem( sitem, "alarm" );
				QString a = i->text( 1 );
				if ( a.contains( ' ' ) )
					a.remove( 0, a.find( ' ' ) );
				else
					a = "";
				a.prepend( m.dat[ 1 ] & sns_flg_aln ? "min" : ( m.dat[ 1 ] & sns_flg_alx ? "max" : "off" ) );
				i->setText( 1, a );
				
				i = sitem->firstChild();
				while ( i && i->text( 0 ) != "range" )
					i = i->nextSibling();
				if ( ! i )
					i = new QListViewItem( sitem, "range" );
				i->setText( 1, QString( "( %1 ~ %2 )" ).arg( m.dat[ 2 ], 0, 16 ).arg( m.dat[ 3 ], 0, 16 ) );
			}
			
		}
		break;
	case ( cmd_sns + 8 ):
	case ( cmd_sns + 9 ):
	case ( cmd_sns + 10 ):
	case ( cmd_sns + 11 ):
	case ( cmd_sns + 12 ):
	case ( cmd_sns + 13 ):
	case ( cmd_sns + 14 ):
	case ( cmd_sns + 15 ):
		if ( item )
		{
			QListViewItem * sitem = item->firstChild();
			while ( sitem && sitem->text( 0 ) != "sns" )
				sitem = sitem->nextSibling();
			if ( sitem )
				sitem = sitem->firstChild();
			while ( sitem && sitem->text( IDCOL ) != QString( "%1" ).arg( m.cmd - cmd_sns - 8, 0, 16 ) )
				sitem = sitem->nextSibling();
				
			
			if ( sitem )
			{
				QListViewItem * i = sitem->firstChild();
				while ( i && i->text( 0 ) != "alarm" )
					i = i->nextSibling();
				if ( ! i )
					i = new QListViewItem( sitem, "alarm" );
				QString a = i->text( 1 );
				if ( a.contains( ' ' ) )
					a = a.left( a.find( ' ' ) );
				a.append( QString( " ( %1 ~ %2 )" ).arg( m.dat[ 2 ], 0, 16 ).arg( m.dat[ 3 ], 0, 16 ) );
				i->setText( 1, a );
				
				sitem->setText( UNITCOL, QString::number( m.dat[ 1 ] ) );
				sitem->setText( OFSCOL, QString::number( m.dat[ 0 ] ) );
			}
		}
		break;
	case ( cmd_trg + 0 ):
	case ( cmd_trg + 1 ):
	case ( cmd_trg + 2 ):
	case ( cmd_trg + 3 ):
	case ( cmd_trg + 4 ):
	case ( cmd_trg + 5 ):
	case ( cmd_trg + 6 ):
	case ( cmd_trg + 7 ):
		if ( item )
		{
			QListViewItem * titem = item->firstChild();
			while ( titem && titem->text( 0 ) != "trg" )
				titem = titem->nextSibling();
			if ( titem )
				titem = titem->firstChild();
			while ( titem && titem->text( IDCOL ) != QString( "%1" ).arg( ( m.cmd - cmd_trg ), 0, 16 ) )
				titem = titem->nextSibling();
			
			if ( !titem )
			{
				titem = new QListViewItem( item, "trg" );
				titem->setOpen( true );
			}
			if ( titem )
			{
				QString tname;
				for ( int i = 0; i < 8; i++ )
					tname += QChar( m.dat[ i ] );
				
				titem->setText( 0, tname );
				titem->setText( 1, ( m.par & 0x01 ? "on" : "off" ) );
			}
		}
		break;
	}
	
	if ( ! item )
	{
		msg	n;
		n.id	= m.id;
		n.ide	= true;
		n.cmd	= cmd_prp;
		n.par	= 0x00;
		n.rtr	= true;
		n.len	= 0;
		snd( n );
	}
	}
	
	new QListViewItem( messages, t, name + " (" + idd + ")", cmd, par, req, dat, ide );
	while ( messages->childCount() > max_messages )
	{
		QListViewItem * lvi = 0;
		
		for ( QListViewItem * ivl = messages->firstChild(); ivl; ivl = ivl->nextSibling() )
		{
			if ( ! lvi || ivl->text( 0 ) < lvi->text( 0 ) )
				lvi = ivl;
		}
		
		delete lvi;
	}
}

void	CAN::snd( msg m )
{
	sndlist.append( m );
}

void	CAN::sndtimer()
{
	if ( sndlist.isEmpty() )
		return;
	
	emit	sigSnd( sndlist.first() );
	
	sndlist.pop_front();
}
