#include <qfiledialog.h>
#include <qhbuttongroup.h>
#include <qlabel.h>
#include <qlineedit.h>
#include <qprogressbar.h>
#include <qpushbutton.h>
#include <qtimer.h>

#include "eeprom.h"
#include "byteview.h"

#define tout 200

eeprom::eeprom( QWidget * parent )
	: QVGroupBox( parent )
{
	setTitle( "eeprom" );
	
	file = new QLineEdit( this );
	file->setText( "/home/veldrin/field13/server/html/html.dat" );
	
	acts = new QHButtonGroup( this );
	acts->setFrameShape( QFrame::NoFrame );
	acts->setMargin( 0 );
	
	prog = new QProgressBar( this );
	prog->setTotalSteps( 65536 );
	
	stat = new QLabel( this );
	stat->setText( "ready" );
	
	new QPushButton( "read", acts );
	new QPushButton( "verify", acts );
	new QPushButton( "write", acts );
	new QPushButton( "view", acts );
	new QPushButton( "browse", acts );
	
	connect( acts, SIGNAL( clicked( int ) ), this, SLOT( action( int ) ) );
	
	bview = new byteview( 0 );
	
	tim = new QTimer( this );
	connect( tim, SIGNAL( timeout() ), this, SLOT( actionPro() ) );
}

void eeprom::action( int id )
{
	//qDebug( QString::number( id ) );
	
	switch ( (actions) id )
	{
	case abort:
		{
			actn = abort;
			prog->setProgress( 0 );
			trys = 0;
			tim->stop();
			acts->setEnabled( true );
			file->setEnabled( true );
		}
		break;
	case ready:
		{
			stat->setText( "complete" );
			actn = ready;
			prog->setProgress( prog->totalSteps() );
			trys = 0;
			tim->stop();
			acts->setEnabled( true );
			file->setEnabled( true );
		}
		break;
	case view:
		{
			QFile fl( file->text() );
			
			if ( fl.open( IO_ReadOnly ) )
			{
				int c = 0;
				while ( c < 65536*2 )
				{
					int x = fl.getch();
					if ( x < 0 )
						break;
					bview->buffer[ c++ ] = (byte) ( x & 0xff );
				}
				while ( c < 65536*2 )
				{
					bview->buffer[ c++ ] = 0xff;
				}
					
				fl.close();
				bview->updateContents();
				bview->show();
			}
			else
			{
				stat->setText( "failed to open file" );
			}
		}
		break;
		
	case browse:
		{
			QString fn = QFileDialog::getSaveFileName( "/home/veldrin/field13/server/html/html.dat", "*", this, "", "select file" );
			if ( ! fn.isEmpty() )
			{
				file->setText( fn );
			}
		}
		break;
		
	case read:
		{
			if ( file->text().isEmpty() )
				stat->setText( "file is not" );
			else if ( QFile::exists( file->text() ) )
				stat->setText( "file exists" );
			else
			{
				actn = (actions) id;
				prog->setProgress( 0 );
				acts->setEnabled( false );
				file->setEnabled( false );
				tim->start( tout, true );
			}
		}
		break;
		
	case very:
		stat->setText( "verify ... " );
	case write:
		{
			QFile fl( file->text() );
			
			if ( fl.open( IO_ReadOnly ) )
			{
				int c = 0;
				while ( c < 65536*2 )
				{
					int x = fl.getch();
					if ( x < 0 )
						break;
					buffer[ c++ ] = (byte) ( x & 0xff );
				}
				while ( c < 65536*2 )
				{
					buffer[ c++ ] = 0xff;
				}
					
				fl.close();
				
				actn = (actions) id;
				prog->setProgress( 0 );
				acts->setEnabled( false );
				file->setEnabled( false );
				tim->start( tout, true );
			}
			else
			{
				stat->setText( "failed to open file" );
			}
		}
		break;

	default:
		break;
	}
}

void eeprom::actionPro()
{
	tim->stop();
	
	switch ( actn )
	{
	case read: // read
		{
			if ( trys == 0 )
				stat->setText( QString( "reading page %1 of %1" ).arg( prog->progress() / 64 ).arg( 65536 / 64 ) );
			if ( trys++ > 5 )
			{
				stat->setText( stat->text() + " timeout" );
				action( abort );
			}
			byte tb[ 3 ];
			tb[ 0 ] = 0x81;
			tb[ 1 ] = ( prog->progress() & 0xff00 ) >> 8;
			tb[ 2 ] =   prog->progress() & 0x00ff;
			emit transmit( tb, 3 );
			
			tim->start( tout, true );
		}
		break;
	case very:
		{
			if ( trys++ > 5 )
			{
				stat->setText( stat->text() + " timeout" );
				action( abort );
			}
			else
			{
				byte tb[ 3 ];
				tb[ 0 ] = 0x81;
				tb[ 1 ] = ( prog->progress() & 0xff00 ) >> 8;
				tb[ 2 ] =   prog->progress() & 0x00ff;
				emit transmit( tb, 3 );
				tim->start( tout, true );
			}
		}
		break;
	case write:
		{
			if ( trys == 0 )
				stat->setText( QString( "writing page %1 of %1" ).arg( prog->progress() / 64 ).arg( 65536 / 64 ) );
			if ( trys++ > 3 )
			{
				stat->setText( stat->text() + " timeout" );
				action( abort );
			}
			else
			{
				byte tb[ 400 ];
				tb[ 0 ] = 0x82;			
				tb[ 1 ] = ( prog->progress() >> 8 ) & 0xff;
				tb[ 2 ] =   prog->progress() & 0xff;
				for ( int i = 0; i < 128; i++ )
				{
					tb[ 3 + i ] = buffer[ prog->progress() * 2 + i ];
				}
				emit transmit( tb, 3 + 128 );
				tim->start( tout * 10, true );
			}
		}
		break;
	default:
		acts->setEnabled( true );
		break;
	}
}

void eeprom::receive( byte * rb, int len )
{
	switch ( actn )
	{
	case write:
		if ( rb[ 0 ] == 0x82 )
		{
			int o = rb[ 2 ];
			o += rb[ 1 ] << 8;
			if ( o == prog->progress() )
			{
				if ( prog->progress() + ( len - 3 ) / 2 >= prog->totalSteps() )
				{
					action( ready );
				}
				else
				{
					prog->setProgress( o + ( len - 3 ) / 2 );
					trys = 0;
					actionPro();
				}
			}
		}
		break;
	case very:
		{
			int o = rb[ 2 ];
			o += rb[ 1 ] << 8;
			bool match = true;
			for ( int c = 0; ( c < len - 3 ) && ( o * 2 + c < 65536 * 2 ); c++ )
				if ( buffer[ o * 2 + c ] != rb[ c + 3 ] )
					match = false;
			if ( ! match )
			{
				stat->setText( stat->text() + "failed" );
				action( abort );
			}
			else if ( o == prog->progress() )
			{
				if ( prog->progress() + ( len - 3 ) / 2 >= prog->totalSteps() )
				{
					action( ready );
				}
				else
				{
					prog->setProgress( prog->progress() + ( len - 3 ) / 2 );
					trys = 0;
					actionPro();
				}
			}
		}
		break;
	case read:
		{
			int o = rb[ 2 ];
			o += rb[ 1 ] << 8;
			o *= 2;
			int l = len - 3; 
			l /= 2;
			ASSERT( l == 64 );
			for ( int c = 0; ( c < len - 3 ) && ( o + c < 65536 * 2 ); c++ )
				buffer[ o + c ] = rb[ c + 3 ];
			if ( o == prog->progress() * 2 )
			{
				if ( prog->progress() + l >= prog->totalSteps() )
				{
					QFile fl( file->text() );
					try
					{
						if ( !fl.open( IO_WriteOnly ) )
							throw int( 0 );
						
						for ( int i = 0; i < 65536 * 2; i++ )
						{
							if ( fl.putch( buffer[ i ] ) == -1 )
								throw int( 0 );
						}
						fl.close();
						action( ready );
					}
					catch ( int )
					{
						stat->setText( "failed to write file" );
						action( abort );
					}
				}
				else
				{
					prog->setProgress( prog->progress() + l );
					trys = 0;
					actionPro();
				}
			}
		}
		break;
	default:
		break;
	}
}
