#include "uss.h"

#include <qpainter.h>
#include <qpushbutton.h>

USS::USS( QWidget * parent )
	: QVBox( parent )
{
	graf = new QWidget( this );
	graf->setFixedSize( 512, 256 );
	graf->installEventFilter( this );
	graf->setBackgroundMode( NoBackground );
	
	QPushButton * s = new QPushButton( this );
	s->setText( "scan" );
	connect( s, SIGNAL( clicked() ), this, SLOT( start() ) );
	
	x = 0;
	
	for ( int y = 0; y < 512; y++ )
	{
		data[ y ] = 0;
	}
}

void USS::start()
{
	msg m;
	m.id = 0x06;
	m.cmd = 0x23;
	m.par = 0x00;
	m.rtr = false;
	m.ide = true;
	m.len = false;
	
	emit sigSnd( m );
	
	x = 0;
	
	for ( int y = 0; y < 512; y++ )
	{
		data[ y ] = 0;
	}
	
	graf->update();
}

void USS::rcv( msg m )
{
	if ( m.cmd != 0x23 )
		return;
	
	if ( m.len == 0 )
	{
		//qDebug( "uss ready" );
		x = 0;
		m.id = 0x06;
		m.cmd = 0x23;
		m.par = 0;
		m.ide = true;
		m.len = 8;
		m.rtr = false;
		
		emit sigSnd( m );
	}
	else
	{
		//qDebug( "uss data" );
		for ( int y = 0; y < m.len && y < 512; y++ )
		{
			data[ m.par*2 + y ] = m.dat[ y ];
		}
		graf->update();
		
		if ( x < m.par*2 + m.len )
		{
			x = m.par*2 + m.len;
		}
		if ( x >= 512 )
			x = 0;
		else
		{
			msg m;
			m.id = 0x06;
			m.cmd = 0x23;
			m.par = x >> 1;
			m.ide = true;
			m.len = 8;
			m.rtr = false;
			emit sigSnd( m );
		}
	}
}

bool USS::eventFilter( QObject * o, QEvent * e )
{
	if ( o != graf )
		return false;
	
	if ( e->type() != QEvent::Paint )
		return false;
	
	QPixmap q( QSize( 512, 256 ) );
	
	QPainter p;
	if ( ! q.isNull() )
		p.begin( &q );
	else
		p.begin( graf );
		
	p.fillRect( graf->rect(), graf->backgroundColor() );
	
	p.setPen( QPen( QColor( "red" ), 1 ) );
	
	unsigned int m = 0;
	
	for ( int y = 0; y < 511; y++ )
	{
		m += data[ y ];
		p.drawLine( QPoint( y, 255-data[ y ] ), QPoint( y+1, 255-data[ y + 1 ] ) );
	}
	
	m /= 512;
	
	p.setPen( QPen( QColor( "blue" ), 1, DotLine ) );
	p.drawLine( QPoint( 0, 255-m*2 ), QPoint( 511, 255-m*2 ) );
	
	
	if ( ! q.isNull() )
		bitBlt( graf, graf->rect().topLeft(), &q, q.rect() );
}
