
; *
; *  manages up to 8 sensors with min/max value saving & adjustable min/max alarm
; *

;.equ	sns_dat	= 0x200	; sram offset for sensor data
;.equ	sns_num	= 1	; total number of sensors
.equ	sns_num_max	= 8
.equ	sns_nxt	= 0x10	; sensor data record size
.equ	sns_dat_end	= ( sns_dat + sns_num * sns_nxt )
;.equ	sns_cfg	= 0x000	; eeprom offset for sensor config

.equ	sns_val	= 0x000	; sensor value
.equ	sns_flg	= 0x001	; sensor flags
.equ	sns_min	= 0x002	; minimum value
.equ	sns_max	= 0x003	; maximum value
.equ	sns_id0	= 0x004	; sensor name
.equ	sns_id1	= 0x005
.equ	sns_id2	= 0x006
.equ	sns_id3	= 0x007

.equ	sns_ofs	= 0x008	; offset value
.equ	sns_unt	= 0x009	; unity code
.equ	sns_aln	= 0x00a	; minimum alarm value
.equ	sns_alx	= 0x00b	; maximum alarm value
.equ	sns_trs	= 0x00c	; alarm treshold
.equ	sns_mux	= 0x00d	; ADMUX value (ADC Channel)
.equ	sns_raw	= 0x00e	; raw sensor value

.equ	sns_flg_aln	= 0	; min alarm flag
.equ	sns_flg_alx	= 1	; max alarm flag
.equ	sns_flg_psa	= 4	; process & send if alarm triggered
.equ	sns_flg_psc	= 5	; process & send if sensor value changed
.equ	sns_flg_psr	= 6	; process & send if min/max range extended

.equ	sns_unt_raw	= 0x00	; raw value
.equ	sns_unt_v16	= 0x01	; Voltage*16/100
.equ	sns_unt_amp	= 0x02	; Amps/100 ( Rsns 0.1 Gain 10 )
.equ	sns_unt_cel	= 0x03	; Degrees Celsius
.equ	sns_unt_16m	= 0x04	; Meters*16
.equ	sns_unt_amq	= 0x05	; Amps/50  ( Rsns 0.05 Gain 10 )


.include "sensor.inc"

.cseg


; *
; *  init sensor data

msr_init:	
	sety	sns_dat
	
	ldi	r16,	sns_num
msr_init1:
	rcall	msr_clr_chx
	adiw	YL,	sns_nxt
	dec	r16
	brne	msr_init1
	
	;rjmp	msr_load


; *
; *  load sensor configuration data

msr_load:
	in	r16,	SREG
	push	r16

	cli

	; copy sensor config
	setz	sns_cfg
	sety	sns_dat

	ldi	dh,	sns_num
msr_load_1:
	rcall	ee_read
	std	Y+sns_id0,	dl
	rcall	ee_read
	std	Y+sns_id1,	dl
	rcall	ee_read
	std	Y+sns_id2,	dl
	rcall	ee_read
	std	Y+sns_id3,	dl
	rcall	ee_read
	std	Y+sns_mux,	dl
	rcall	ee_read
	andi	dl,	( 1 << sns_flg_psa ) | ( 1 << sns_flg_psc ) | ( 1 << sns_flg_psr )
	std	Y+sns_flg,	dl
	rcall	ee_read
	std	Y+sns_ofs,	dl
	rcall	ee_read
	std	Y+sns_unt,	dl
	rcall	ee_read
	std	Y+sns_aln,	dl
	rcall	ee_read
	std	Y+sns_alx,	dl
	rcall	ee_read
	std	Y+sns_trs,	dl

	adiw	YL,	sns_nxt
	dec	dh
	brne	msr_load_1

	pop	r16
	out	SREG,	r16
	ret


; *
; *  measure all sensors

msr:
	push	dl
	push	dh
	pushy

	; enable AD Converter
	ldi	dh,	( 1 << ADEN ) + ( 0 << ADPS2 ) + ( 1 << ADPS1 ) + ( 1 << ADPS0 )
	out	ADCSRA,	dh

	; load pointer for data storage
	sety	sns_dat
	
	clr	dh
msr_1:
	; set mux register
	ldd	dl,	Y+sns_mux
	tst	dl
	breq	msr_4
	out	ADMUX,	dl

	; clear result buffer	
	clr	dl

	push	dh

msr_2:
	; start ad conversion
	sbi	ADCSRA,	ADSC
msr_3:
	; wait for conversion to finish
	sbic	ADCSRA,	ADSC
	rjmp	msr_3
	
	; and wait for two equal results in a row
	mov	dh,	dl
	in	dl,	ADCH
	cpse	dh,	dl
	rjmp	msr_2
	
	pop	dh

	; save new value
	rcall	msr_ins_val
	
msr_4:
	adiw	YL,	sns_nxt
	
	inc	dh
	cpi	dh,	sns_num
	brcs	msr_1

	; disable AD Converter
	clr	dl
	out	ADCSRA,	dl

	popy
	pop	dh
	pop	dl

	ret


; *
; *  insert sensor value & update min/max/alarm

msr_ins_val:
	push	r16

	pushz
	setz	sns_clc_tab
	clr	r16
	andi	dh,	0b111
	add	ZL,	dh
	adc	ZH,	r16
	icall
	popz
	
	; check if result differs from last value
	ld	r16,	Y
	cp	r16,	dl
	brne	msr_chx_new
	
	pop	r16
	ret

msr_chx_new:
	; store result
	st	Y,	dl
	
	push	r17
	push	r18

	; clear snd status flag
	clr	r18
	
	; check value change flag
	ldd	r16,	Y+sns_flg
	sbrc	r16,	sns_flg_psc
	sbr	r18,	( 1 << sns_flg_psc )

	; update min/max values
	ldd	r17,	Y+sns_min
	cp	dl,	r17
	brcc	msr_chx_nmin
	std	Y+sns_min,	dl
	sbrc	r16,	sns_flg_psr
	sbr	r18,	( 1 << sns_flg_psr )
msr_chx_nmin:
	ldd	r17,	Y+sns_max
	cp	r17,	dl
	brcc	msr_chx_nmax
	std	Y+sns_max,	dl
	sbrc	r16,	sns_flg_psr
	sbr	r18,	( 1 << sns_flg_psr )
msr_chx_nmax:

	; value below min alarm ?
	ldd	r16,	Y+sns_aln
	cp	dl,	r16
	brcc	msr_chx_naln

	; min alarm ! - flag already set ?
	ldd	r16,	Y+sns_flg
	sbrc	r16,	sns_flg_aln
	rjmp	msr_chx_alx		; yes - no change

	; set min alarm flag
	sbr	r16,	( 1 << sns_flg_aln )
	std	Y+sns_flg,	r16
	; check psa flag
	sbrc	r16,	sns_flg_psa
	sbr	r18,	( 1 << sns_flg_psa )
	rjmp	msr_chx_alx

msr_chx_naln:
	; value above min alarm + threshold ?
	ldd	r17,	Y+sns_trs
	add	r16,	r17
	cp	dl,	r16
	brcs	msr_chx_alx
	
	; check min alarm flag
	ldd	r16,	Y+sns_flg
	sbrs	r16,	sns_flg_aln
	rjmp	msr_chx_alx
	
	; clear min alarm flag
	cbr	r16,	( 1 << sns_flg_aln )
	std	Y+sns_flg,	r16
	; check psa flag
	sbrc	r16,	sns_flg_psa
	sbr	r18,	( 1 << sns_flg_psa )

	; check max alarm
msr_chx_alx:
	; value above max alarm
	ldd	r16,	Y+sns_alx
	cp	r16,	dl
	brcc	msr_chx_nalx

	; max alarm ! - flag already set ?
	ldd	r16,	Y+sns_flg
	sbrc	r16,	sns_flg_alx
	rjmp	msr_chx_fin		; yes - no change - exit

	; set max alarm flag
	sbr	r16,	( 1 << sns_flg_alx )
	std	Y+sns_flg,	r16
	; check psa flag
	sbrc	r16,	sns_flg_psa
	sbr	r18,	( 1 << sns_flg_psa )
	rjmp	msr_chx_fin
	
msr_chx_nalx:
	; value below maximum threshold ?
	ldd	r17,	Y+sns_trs
	sub	r16,	r17
	cp	dl,	r16
	brcc	msr_chx_fin		; no - exit

	; check max alarm flag
	ldd	r16,	Y+sns_flg
	sbrs	r16,	sns_flg_alx
	rjmp	msr_chx_fin
	
	; reset alarm flag
	cbr	r16,	( 1 << sns_flg_alx )
	std	Y+sns_flg,	r16
	; check psa flag
	sbrc	r16,	sns_flg_psa
	sbr	r18,	( 1 << sns_flg_psa )
	
msr_chx_fin:
	tst	r18
	breq	msr_chx_skip_snd
	
	mov	dl,	r18
	rcall	msr_can_snd
	rcall	msr_pro

msr_chx_skip_snd:
	pop	r18
	pop	r17
	pop	r16
	ret


; *
; *  send sensor channel command through can interface

msr_can_snd:
	push	dl
	push	dh
	pushy
	pushz
	push	dl
	
	bst	dh,	3
	andi	dh,	0x07

	clr	dl	
	swap	dh

	sety	sns_dat
	add	YL,	dh
	adc	YH,	dl
	
	swap	dh
	
	setz	can_dat
	ldi	dl,	can_addr
	st	Z,	dl
	ldi	dl,	can_addr_ext
	st	Z+1,	dl
	ldi	dl,	cmd_sns_0
	bld	dl,	3
	add	dl,	dh
	std	Z+2,	dl
	pop	dl
	std	Z+3,	dl
	
	adiw	ZL,	4
	ldi	dh,	8
	st	Z+,	dh
	
	brtc	msr_can_snd_1
	adiw	YL,	8
	
msr_can_snd_1:
	ld	dl,	Y+
	st	Z+,	dl
	dec	dh
	brne	msr_can_snd_1
	
	sety	can_dat
	rcall	can_snd_y
	
	popz
	popy
	pop	dh
	pop	dl
	ret


; *
; *  execute sensor event handler procedure

msr_pro:
	pushz
	setz	sns_prc_tab
	
	andi	dh,	0b00000111
	add	ZL,	dh
	brcc	msr_pro_noc
	inc	ZH
msr_pro_noc:
	icall
	
	popz
	ret


; *
; *  clear sensor channel data

msr_clr_chx:
	clr	r0
	st	Y,	r0
	std	Y+sns_flg,	r0
	std	Y+sns_max,	r0
	dec	r0
	std	Y+sns_min,	r0
	ret


; *
; *  save sensor configuration data

msr_save:
	in	r16,	SREG
	push	r16

	cli

	; save sensor config
	setz	sns_cfg
	sety	sns_dat
	ldi	dh,	sns_num
msr_save_1:
	ldd	dl,	Y+sns_id0
	rcall	ee_write
	ldd	dl,	Y+sns_id1
	rcall	ee_write
	ldd	dl,	Y+sns_id2
	rcall	ee_write
	ldd	dl,	Y+sns_id3
	rcall	ee_write
	ldd	dl,	Y+sns_mux
	rcall	ee_write
	ldd	dl,	Y+sns_flg
	andi	dl,	( 1 << sns_flg_psa ) | ( 1 << sns_flg_psc ) | ( 1 << sns_flg_psr )
	rcall	ee_write
	ldd	dl,	Y+sns_ofs
	rcall	ee_write
	ldd	dl,	Y+sns_unt
	rcall	ee_write
	ldd	dl,	Y+sns_aln
	rcall	ee_write
	ldd	dl,	Y+sns_alx
	rcall	ee_write
	ldd	dl,	Y+sns_trs
	rcall	ee_write

	adiw	YL,	sns_nxt
	dec	dh
	brne	msr_save_1

	pop	r16
	out	SREG,	r16
	ret


; *
; *  get sensor value ( dh7:4 - sensor channel, dh3:0 sensor data offset )

msr_get_sns_val:
	pushy
	sety	sns_dat
	add	YL,	dh
	brcc	msr_get_sns_val_noc
	inc	YH
msr_get_sns_val_noc:
	ld	dl,	Y
	popy
	ret


