
; tcp header
.equ	TCP_SRCPRT	=	IP_DAT + 0x00
.equ	TCP_DSTPRT	=	IP_DAT + 0x02
.equ	TCP_SEQ	=	IP_DAT + 0x04
.equ	TCP_ACK	=	IP_DAT + 0x08
.equ	TCP_HDR	=	IP_DAT + 0x0c
.equ	TCP_FLG	=	IP_DAT + 0x0d
.equ	TCP_WIN	=	IP_DAT + 0x0e
.equ	TCP_CHK	=	IP_DAT + 0x10
.equ	TCP_PTR	=	IP_DAT + 0x12
.equ	TCP_DAT	=	IP_DAT + 0x14

; tcp ports
.equ	TCP_PRT_HTTP	=	80
;.equ	TCP_PRT_TELNET	=	23
;.equ	TCP_PRT_FTP	=	21

; tcp flags
.equ	TCP_FLG_FIN	=	0
.equ	TCP_FLG_SYN	=	1
.equ	TCP_FLG_RST	=	2
.equ	TCP_FLG_PSH	=	3
.equ	TCP_FLG_ACK	=	4


; tcp connection data
;.equ	tcp_sav	=	0x300
.equ	tcp_sav_num	=	0x08	; number of tcp connections
.equ	tcp_sav_nxt	=	0x20	; record size
.equ	tcp_sav_end	=	( tcp_sav + ( tcp_sav_num * tcp_sav_nxt ) )

.equ	tcp_sav_prt	=	0x00	; remote port
.equ	tcp_sav_adr	=	0x02	; remote address
.equ	tcp_sav_sts	=	0x06	; connection state
.equ	tcp_sav_tim	=	0x07	; timeout counter
.equ	tcp_sav_seq	=	0x08	; sequence number
.equ	tcp_sav_ack	=	0x0c	; acknoledge number
.equ	tcp_sav_nsq	=	0x10	; next sequence number
.equ	tcp_sav_dat	=	0x14	; application data

; tcp connection states
.equ	tcp_sts_open	=	0	; no connection
.equ	tcp_sts_syn	=	1	; connection in use


; *
; *  clear connection data

tcp_init:
	sety	tcp_sav
	clr	r0
	clr	r16
tcp_init_1:
	st	Y+,	r0
	dec	r16
	brne	tcp_init_1
	ret


; *
; *  get pointer to tcp packet data

tcp_dat_ptr:
	sety	nic_dat
	ldd	r16,	Y+TCP_HDR
	andi	r16,	0xf0
	lsr	r16
	lsr	r16
	clr	r0
	add	YL,	r16
	adc	YH,	r0
	adiw	YL,	20
	ret


; *
; *  calculate tcp data length

tcp_dat_len:
	sety	nic_dat
	ldd	XH,	Y+IP_LEN+0
	ldd	XL,	Y+IP_LEN+1
	ldd	r16,	Y+TCP_HDR
	andi	r16,	0xf0
	lsr	r16
	lsr	r16
	sub	XL,	r16
	sbci	XH,	0
	subi	XL,	20
	sbci	XH,	0
	ret


; *
; *  process incoming tcp packets

tcp_in:	
	sety	nic_dat
	; check destination port
	ldd	dh,	Y+TCP_DSTPRT
	tst	dh
	brne	tcp_drop_port
	ldd	dl,	Y+TCP_DSTPRT+1
	cpi	dl,	TCP_PRT_HTTP
	brne	tcp_drop_port
	rjmp	http_in
tcp_drop_port:
	rjmp	icmp_dest_port
	

; *
; *  return tcp reset packet

tcp_reset:
	sety	nic_dat
	
	; set tcp flags
	ldi	r16,	( 1 << TCP_FLG_RST ) | ( 1 << TCP_FLG_ACK )
	std	Y+TCP_FLG,	r16
	ldi	r16,	0x50
	std	Y+TCP_HDR,	r16
	
	; set ip total length
	clr	r16
	std	Y+IP_LEN+0,	r16
	ldi	r16,	(TCP_DAT-ETH_DAT)
	std	Y+IP_LEN+1,	r16
	
	; swap ports
	ldd	dh,	Y+TCP_SRCPRT
	ldd	dl,	Y+TCP_DSTPRT
	std	Y+TCP_SRCPRT,	dl
	std	Y+TCP_DSTPRT,	dh
	ldd	dh,	Y+TCP_SRCPRT+1
	ldd	dl,	Y+TCP_DSTPRT+1
	std	Y+TCP_SRCPRT+1,	dl
	std	Y+TCP_DSTPRT+1,	dh
	
	; swap seq / ack num
	ldd	dh,	Y+TCP_SEQ+0
	ldd	dl,	Y+TCP_ACK+0
	std	Y+TCP_SEQ+0,	dl
	std	Y+TCP_ACK+0,	dh
	ldd	dh,	Y+TCP_SEQ+1
	ldd	dl,	Y+TCP_ACK+1
	std	Y+TCP_SEQ+1,	dl
	std	Y+TCP_ACK+1,	dh
	ldd	dh,	Y+TCP_SEQ+2
	ldd	dl,	Y+TCP_ACK+2
	std	Y+TCP_SEQ+2,	dl
	std	Y+TCP_ACK+2,	dh
	ldd	dh,	Y+TCP_SEQ+3
	ldd	dl,	Y+TCP_ACK+3
	std	Y+TCP_SEQ+3,	dl
	std	Y+TCP_ACK+3,	dh
	
	; clear checksum
	clr	dl
	clr	dh
	std	Y+TCP_CHK+0,	dh
	std	Y+TCP_CHK+1,	dl
	
	; set new checksum
	rcall	tcp_chksum
	sety	nic_dat
	std	Y+TCP_CHK+0,	dh
	std	Y+TCP_CHK+1,	dl

	; make ip header
	rcall	ip_header
	
	; send packet
	rjmp	ip_out


; *
; *  check if this ack is the first one after syn ( ack num = 1 )

tcp_chk_fst_ack:
	push	dl
	ldd	dl,	Y+TCP_ACK+0
	tst	dl
	brne	tcp_chk_fst_ack_fin
	ldd	dl,	Y+TCP_ACK+1
	tst	dl
	brne	tcp_chk_fst_ack_fin
	ldd	dl,	Y+TCP_ACK+2
	tst	dl
	brne	tcp_chk_fst_ack_fin
	ldd	dl,	Y+TCP_ACK+3
	cpi	dl,	1
tcp_chk_fst_ack_fin:
	pop	dl
	ret


; *
; *  free connection buffer

tcp_sav_clr:
	clr	r16
	std	Z+tcp_sav_sts,	r16
	ret


; *
; *  copy acknoledge number

tcp_sav_ack_cpy:
	ldd	r16,	Y+TCP_SEQ+0
	std	Z+tcp_sav_ack+0,	r16
	ldd	r16,	Y+TCP_SEQ+1
	std	Z+tcp_sav_ack+1,	r16
	ldd	r16,	Y+TCP_SEQ+2
	std	Z+tcp_sav_ack+2,	r16
	ldd	r16,	Y+TCP_SEQ+3
	std	Z+tcp_sav_ack+3,	r16
	ret


; *
; *  add XH:XL to ack number

tcp_sav_ack_add:
	ldd	r1,	Z+tcp_sav_ack+0
	ldd	r2,	Z+tcp_sav_ack+1
	ldd	r3,	Z+tcp_sav_ack+2
	ldd	r4,	Z+tcp_sav_ack+3
	
	clr	r0
	add	r4,	XL
	adc	r3,	XH
	adc	r2,	r0
	adc	r1,	r0
	adc	r4,	r0
	
	std	Z+tcp_sav_ack+0,	r1
	std	Z+tcp_sav_ack+1,	r2
	std	Z+tcp_sav_ack+2,	r3
	std	Z+tcp_sav_ack+3,	r4
	ret


; *
; *  check if next sequence number matches
	
tcp_sav_nsq_chk:
	ldd	dl,	Z+tcp_sav_nsq+3
	ldd	dh,	Y+TCP_ACK+3
	cpse	dh,	dl
	ret
	ldd	dl,	Z+tcp_sav_nsq+2
	ldd	dh,	Y+TCP_ACK+2
	cpse	dh,	dl
	ret
	ldd	dl,	Z+tcp_sav_nsq+1
	ldd	dh,	Y+TCP_ACK+1
	cpse	dh,	dl
	ret
	ldd	dl,	Z+tcp_sav_nsq+0
	ldd	dh,	Y+TCP_ACK+0
	cp	dh,	dl
	ret


; *
; *  check if old sequence number matches

tcp_sav_seq_chk:
	ldd	dl,	Z+tcp_sav_seq+3
	ldd	dh,	Y+TCP_ACK+3
	cpse	dh,	dl
	ret
	ldd	dl,	Z+tcp_sav_seq+2
	ldd	dh,	Y+TCP_ACK+2
	cpse	dh,	dl
	ret
	ldd	dl,	Z+tcp_sav_seq+1
	ldd	dh,	Y+TCP_ACK+1
	cpse	dh,	dl
	ret
	ldd	dl,	Z+tcp_sav_seq+0
	ldd	dh,	Y+TCP_ACK+0
	cp	dh,	dl
	ret


; *
; *  copy next sequence number to old sequence number

tcp_sav_seq_cpy:
	ldd	r16,	Z+tcp_sav_nsq+0
	std	Z+tcp_sav_seq+0,	r16
	ldd	r16,	Z+tcp_sav_nsq+1
	std	Z+tcp_sav_seq+1,	r16
	ldd	r16,	Z+tcp_sav_nsq+2
	std	Z+tcp_sav_seq+2,	r16
	ldd	r16,	Z+tcp_sav_nsq+3
	std	Z+tcp_sav_seq+3,	r16
	clr	r16
	std	Z+tcp_sav_tim,	r16
	ret


; *
; *  add XH:XL to next sequence number
	
tcp_sav_nsq_add:
	clr	r0

	ldd	r16,	Z+tcp_sav_seq+3
	add	r16,	XL
	std	Z+tcp_sav_nsq+3,	r16

	ldd	r16,	Z+tcp_sav_seq+2
	adc	r16,	XH
	std	Z+tcp_sav_nsq+2,	r16

	ldd	r16,	Z+tcp_sav_seq+1
	adc	r16,	r0
	std	Z+tcp_sav_nsq+1,	r16

	ldd	r16,	Z+tcp_sav_seq+0
	adc	r16,	r0
	std	Z+tcp_sav_nsq+0,	r16
	ret


; *
; *  search for tcp connection
	
tcp_sav_get:
	sety	nic_dat
	ldd	r0,	Y+TCP_SRCPRT+0
	ldd	r1,	Y+TCP_SRCPRT+1
	ldd	r2,	Y+IP_SRC+0
	ldd	r3,	Y+IP_SRC+1
	ldd	r4,	Y+IP_SRC+2
	ldd	r5,	Y+IP_SRC+3
	
	setz	tcp_sav
tcp_sav_get_loop:
	ldd	r16,	Z+tcp_sav_sts
	tst	r16
	breq	tcp_sav_get_next
	
	ldd	r16,	Z+tcp_sav_prt+0
	cp	r16,	r0
	brne	tcp_sav_get_next
	
	ldd	r16,	Z+tcp_sav_prt+1
	cp	r16,	r1
	brne	tcp_sav_get_next
	
	ldd	r16,	Z+tcp_sav_adr+0
	cp	r16,	r2
	brne	tcp_sav_get_next
	
	ldd	r16,	Z+tcp_sav_adr+1
	cp	r16,	r3
	brne	tcp_sav_get_next

	ldd	r16,	Z+tcp_sav_adr+2
	cp	r16,	r4
	brne	tcp_sav_get_next
	
	ldd	r16,	Z+tcp_sav_adr+3
	cp	r16,	r5
	brne	tcp_sav_get_next
	
	; clear tcp timer
	clr	r16
	st	Z+tcp_sav_tim,	r16
	
	set
	ret
	
tcp_sav_get_next:
	adiw	ZL,	tcp_sav_nxt
	cpi	ZH,	HIGH( tcp_sav_end )
	brcs	tcp_sav_get_loop
	cpi	ZL,	LOW( tcp_sav_end )
	brcs	tcp_sav_get_loop
	
	clt
	ret


; *
; *  open new connection
	
tcp_sav_new:
	setz	tcp_sav
tcp_sav_new_loop:
	ldd	r16,	Z+tcp_sav_sts
	tst	r16
	breq	tcp_sav_new_fill
	
	ldd	r16,	Z+tcp_sav_tim
	cpi	r16,	0xff
	breq	tcp_sav_new_fill
	
	adiw	ZL,	tcp_sav_nxt
	cpi	ZH,	HIGH( tcp_sav_end )
	brcs	tcp_sav_new_loop
	cpi	ZL,	LOW( tcp_sav_end )
	brcs	tcp_sav_new_loop
	
	clt	; too much clients
	ret
	
tcp_sav_new_fill:
	; empty slot found
	sety	nic_dat
	ldd	r16,	Y+TCP_SRCPRT+0
	std	Z+tcp_sav_prt+0,	r16
	ldd	r16,	Y+TCP_SRCPRT+1
	std	Z+tcp_sav_prt+1,	r16
	ldd	r16,	Y+IP_SRC+0
	std	Z+tcp_sav_adr+0,	r16
	ldd	r16,	Y+IP_SRC+1
	std	Z+tcp_sav_adr+1,	r16
	ldd	r16,	Y+IP_SRC+2
	std	Z+tcp_sav_adr+2,	r16
	ldd	r16,	Y+IP_SRC+3
	std	Z+tcp_sav_adr+3,	r16
	clr	r16
	std	Z+tcp_sav_tim  ,	r16
	std	Z+tcp_sav_seq+0,	r16
	std	Z+tcp_sav_seq+1,	r16
	std	Z+tcp_sav_seq+2,	r16
	std	Z+tcp_sav_seq+3,	r16
	std	Z+tcp_sav_nsq+0,	r16
	std	Z+tcp_sav_nsq+1,	r16
	std	Z+tcp_sav_nsq+2,	r16
	std	Z+tcp_sav_dat+0,	r16
	std	Z+tcp_sav_dat+1,	r16
	std	Z+tcp_sav_dat+2,	r16
	std	Z+tcp_sav_dat+3,	r16
	std	Z+tcp_sav_dat+4,	r16
	std	Z+tcp_sav_dat+5,	r16
	ldi	r16,	0x01
	std	Z+tcp_sav_nsq+3,	r16
	std	Z+tcp_sav_sts  ,	r16
	rcall	tcp_sav_ack_cpy
	
	set	; connection saved
	ret
	
	
; *
; *  timeout broken connections

tcp_timer:
	push	r16
	pushz
	setz	tcp_sav
tcp_timer_loop:
	ldd	r16,	Z+tcp_sav_tim
	cpi	r16,	0xff
	breq	tcp_timer_next
	inc	r16
	std	Z+tcp_sav_tim,	r16
tcp_timer_next:
	adiw	ZL,	tcp_sav_nxt
	cpi	ZH,	HIGH( tcp_sav_end )
	brcs	tcp_timer_loop
	cpi	ZL,	LOW( tcp_sav_end )
	brcs	tcp_timer_loop
	
	popz
	pop	r16
	ret
	

; *
; *  setup tcp header with current connection data

tcp_header:
	sety	nic_dat
	
	; set tcp flags
	std	Y+TCP_FLG,	r16
	ldi	r16,	0x50
	std	Y+TCP_HDR,	r16
	
	; set ip total length
	adiw	XL,	(TCP_DAT-ETH_DAT)
	std	Y+IP_LEN+0,	XH
	std	Y+IP_LEN+1,	XL
	
	; swap ports
	ldd	dh,	Y+TCP_SRCPRT
	ldd	dl,	Y+TCP_DSTPRT
	std	Y+TCP_SRCPRT,	dl
	std	Y+TCP_DSTPRT,	dh
	ldd	dh,	Y+TCP_SRCPRT+1
	ldd	dl,	Y+TCP_DSTPRT+1
	std	Y+TCP_SRCPRT+1,	dl
	std	Y+TCP_DSTPRT+1,	dh
	
	; set seq num
	ldd	dl,	Z+tcp_sav_seq+0
	std	Y+TCP_SEQ+0,	dl
	ldd	dl,	Z+tcp_sav_seq+1
	std	Y+TCP_SEQ+1,	dl
	ldd	dl,	Z+tcp_sav_seq+2
	std	Y+TCP_SEQ+2,	dl
	ldd	dl,	Z+tcp_sav_seq+3
	std	Y+TCP_SEQ+3,	dl
	
	; set ack num
	ldd	dl,	Z+tcp_sav_ack+0
	std	Y+TCP_ACK+0,	dl
	ldd	dl,	Z+tcp_sav_ack+1
	std	Y+TCP_ACK+1,	dl
	ldd	dl,	Z+tcp_sav_ack+2
	std	Y+TCP_ACK+2,	dl
	ldd	dl,	Z+tcp_sav_ack+3
	std	Y+TCP_ACK+3,	dl
	
	; clear checksum
	clr	dl
	clr	dh
	std	Y+TCP_CHK+0,	dh
	std	Y+TCP_CHK+1,	dl

	; make ip header
	rjmp	ip_header

	
; *
; *  calculate tcp check sum

tcp_chksum:
	; calculate new checksum
	clr	dl
	clr	dh
	rcall	tcp_chksum_pseudo
	sety	nic_dat+IP_DAT	
	rcall	inet_chksum_part
	ser	r16
	eor	dh,	r16
	eor	dl,	r16
	ret
	

tcp_chksum_pseudo:
	clr	r0
	sety	nic_dat
	; length of tcp message
	ldd	XH,	Y+IP_LEN+0
	ldd	XL,	Y+IP_LEN+1
	subi	XL,	20
	sbci	XH,	0
	; sum up pseudo header
	ldd	r17,	Y+IP_SRC+0
	ldd	r16,	Y+IP_SRC+1
	add	dl,	r16
	adc	dh,	r17
	ldd	r17,	Y+IP_SRC+2
	ldd	r16,	Y+IP_SRC+3
	adc	dl,	r16
	adc	dh,	r17
	ldd	r17,	Y+IP_DST+0
	ldd	r16,	Y+IP_DST+1
	adc	dl,	r16
	adc	dh,	r17
	ldd	r17,	Y+IP_DST+2
	ldd	r16,	Y+IP_DST+3
	adc	dl,	r16
	adc	dh,	r17
	ldd	r16,	Y+IP_PRT
	adc	dl,	r16
	adc	dh,	XH
	adc	dl,	XL
	adc	dh,	r0
	ret
	

