
; udp header
.equ	UDP_SRCPRT	=	IP_DAT + 0x00
.equ	UDP_DSTPRT	=	IP_DAT + 0x02
.equ	UDP_LEN	=	IP_DAT + 0x04
.equ	UDP_CHK	=	IP_DAT + 0x06
.equ	UDP_DAT	=	IP_DAT + 0x08	

; udp control port
.equ	UDP_PRT	=	42135

; udp connection data for can monitoring (only one at a time)
;equ	udp_sav	=	0x200
.equ	udp_sav_end	=	( udp_sav + 0x10 )

; some udp commands
.equ	udp_cmd_stat	=	0x00	; get status
.equ	udp_cmd_can	=	0x31	; send can message
.equ	udp_cmd_read	=	0x81	; read i2c eeprom
.equ	udp_cmd_write	=	0x82	; write i2c eeprom
.equ	udp_cmd_reset	=	0xff	; reset controler


; *
; *  clear connection data

udp_sav_clr:
udp_init:
	clr	r16
	sts	udp_sav+0,	r16
	sts	udp_sav+1,	r16
	ret


; *
; *  store remote host address

udp_save:
	pushx
	pushy
	sety	nic_dat
	
	setx	udp_sav
	ldd	r16,	Y+UDP_SRCPRT+0
	ldd	r17,	Y+UDP_SRCPRT+1
	st	X+,	r16
	st	X+,	r17
	
	ldi	r16,	6
udp_sav_mac:
	ldd	r17,	Y+ETH_SRC
	st	X+,	r17
	adiw	YL,	1
	dec	r16
	brne	udp_sav_mac
	
	subi	YL,	6
	sbci	YH,	0
	
	ldi	r16,	4
udp_sav_ip:
	ldd	r17,	Y+IP_SRC
	st	X+,	r17
	adiw	YL,	1
	dec	r16
	brne	udp_sav_ip
	
	popy
	popx
	ret

	
; *
; *  process incoming udp packets

udp_in:
	sety	nic_dat
	; check destination port
	ldd	r16,	Y+UDP_DSTPRT+0
	cpi	r16,	HIGH( UDP_PRT )
	brne	udp_drop_port
	ldd	r17,	Y+UDP_DSTPRT+1
	cpi	r17,	LOW( UDP_PRT )
	brne	udp_drop_port
	
	; check if checksum is present in header
	ldd	r17,	Y+UDP_CHK+0
	ldd	r16,	Y+UDP_CHK+1
	tst	r16
	brne	udp_chksum_chk
	tst	r17
	breq	udp_chksum_ok
	
udp_chksum_chk:
	; confirm checksum
	rcall	udp_chksum
	cpi	dh,	0xff
	brne	udp_drop
	cpi	dl,	0xff
	brne	udp_drop
	
udp_chksum_ok:
	; check command code
	sety	nic_dat+UDP_DAT
	ld	dl,	Y+
	cpi	dl,	udp_cmd_stat
	breq	udp_stat
	cpi	dl,	udp_cmd_can
	breq	udp_can
	cpi	dl,	udp_cmd_read
	breq	udp_read
	cpi	dl,	udp_cmd_write
	breq	udp_write
	cpi	dl,	udp_cmd_reset
	brne	udp_drop
	rjmp	udp_reset

udp_drop:
	ret
udp_drop_port:
	; send icmp destination port unreachable
	rjmp	icmp_dest_port


; *
; *  return status code 0x01 ( up & running )

udp_stat:
	ldi	dl,	0x1
	st	Y+,	dl
	ldi	XL,	2
	clr	XH
	rjmp	udp_send


; *
; *  insert can message and register remote host as can monitor

udp_can:
	rcall	udp_save
	
	setz	can_dat
	ldi	dh,	0x0d
udp_can_1:
	ld	dl,	Y+
	st	Z+,	dl
	dec	dh
	brne	udp_can_1
	sety	can_dat
	rcall	can_snd_y
	rjmp	can_pro


; *
; *  read and send 128 bytes from i2c eeprom
	
udp_read:
	ld	ZH,	Y+
	ld	ZL,	Y+

	rcall	i2c_read

	ldi	XL,	64
udp_read_loop:
	rcall	i2c_in
	st	Y+,	dh
	st	Y+,	dl
	dec	XL
	brne	udp_read_loop
	
	rcall	i2c_stop
	
	ldi	XL,	131
	clr	XH
	rjmp	udp_send


; *
; *  write and verify 128 bytes

udp_write:
	sbic	I2C_PIN,	I2C_WPRT
	ret
	
	pushy
	ld	ZH,	Y+
	ld	ZL,	Y+
	
	rcall	i2c_write
	
	ldi	XL,	64
udp_write_loop:
	ld	dh,	Y+
	ld	dl,	Y+
	rcall	i2c_out
	dec	XL
	brne	udp_write_loop
	
	rcall	i2c_stop
	
	ldi	r16,	205
	clr	r17
udp_write_wait:
	dec	r17
	brne	udp_write_wait
	dec	r16
	brne	udp_write_wait
	popy
	rjmp	udp_read


; *
; *  reset controller

udp_reset:
	ld	dl,	Y+
	cpi	dl,	0x31
	brne	udp_reset_drop
	ld	dl,	Y+
	cpi	dl,	0xca
	brne	udp_reset_drop
	jmp	0
udp_reset_drop:
	ret


; *
; *  send udp packet back where it came from
	
udp_send:
	sety	nic_dat
	adiw	XL,	( UDP_DAT - IP_DAT )
	std	Y+UDP_LEN+0,	XH
	std	Y+UDP_LEN+1,	XL
	adiw	XL,	20
	std	Y+IP_LEN+0,	XH
	std	Y+IP_LEN+1,	XL

udp_switch:
	sety	nic_dat
	; send udp message back
	ldd	r16,	Y+UDP_SRCPRT+0
	ldd	r17,	Y+UDP_DSTPRT+0
	std	Y+UDP_DSTPRT+0,	r16
	std	Y+UDP_SRCPRT+0,	r17
	ldd	r16,	Y+UDP_SRCPRT+1
	ldd	r17,	Y+UDP_DSTPRT+1
	std	Y+UDP_DSTPRT+1,	r16
	std	Y+UDP_SRCPRT+1,	r17
	
	rcall	ip_header

udp_out:	
	sety	nic_dat
	clr	r16
	std	Y+UDP_CHK+0,	r16
	std	Y+UDP_CHK+1,	r16
	rcall	udp_chksum
	sety	nic_dat
	std	Y+UDP_CHK+0,	dh
	std	Y+UDP_CHK+1,	dl
	
	rjmp	ip_out
	
	
; *
; *  send udp command to connected remote host

udp_snd_cmd:
	setx	udp_sav
	
	; connection established ?
	ld	dh,	X+
	ld	dl,	X+
	tst	dh
	brne	udp_snd_cmd_ok
	tst	dl
	brne	udp_snd_cmd_ok
	
	ret
	
udp_snd_cmd_ok:
	sety	nic_dat
	; set dst port
	std	Y+UDP_DSTPRT+0,	dh
	std	Y+UDP_DSTPRT+1,	dl
	
	; set src port
	ldi	dh,	HIGH( UDP_PRT )
	ldi	dl,	LOW( UDP_PRT )
	std	Y+UDP_SRCPRT+0,	dh
	std	Y+UDP_SRCPRT+1,	dl
	
	; set ethernet type
	ldi	dh,	0x08
	ldi	dl,	0x00
	std	Y+ETH_TYP+0,	dh
	std	Y+ETH_TYP+1,	dl
	
	; set src & dst mac address
	setz	( mymac * 2 )
	ldi	dh,	6
udp_snd1:
	ld	dl,	X+
	std	Y+ETH_DST,	dl
	lpm	dl,	Z+
	std	Y+ETH_SRC,	dl
	adiw	YL,	1
	dec	dh
	brne	udp_snd1
	
	; set src & dst ip address
	sety	nic_dat
	setz	( myip * 2 )
	ldi	dh,	4
udp_snd2:
	ld	dl,	X+
	std	Y+IP_DST,	dl
	lpm	dl,	Z+
	std	Y+IP_SRC,	dl
	adiw	YL,	1
	dec	dh
	brne	udp_snd2
	
	; set ip options
	sety	nic_dat
	ldi	r16,	0x45
	std	Y+IP_VRS,	r16
	clr	r16
	std	Y+IP_TOS,	r16
	std	Y+IP_ID+0,	r16
	std	Y+IP_ID+1,	r16
	std	Y+IP_FRG+0,	r16
	std	Y+IP_FRG+1,	r16
	ldi	r16,	0x40
	std	Y+IP_TTL,	r16
	ldi	r16,	IP_PRT_UDP
	std	Y+IP_PRT,	r16
	
	; set ip len
	ldd	XH,	Y+UDP_LEN+0
	ldd	XL,	Y+UDP_LEN+1
	adiw	XL,	20
	std	Y+IP_LEN+0,	XH
	std	Y+IP_LEN+1,	XL
	
	rjmp	udp_out
	

; *
; *  calculate udp checksum

udp_chksum:
	rcall	tcp_chksum
	; if chksum == 0x0000 then chksum = 0xffff
	tst	dh
	brne	udp_chksum_fin
	tst	dl
	brne	udp_chksum_fin
	ser	dl
	ser	dh
udp_chksum_fin:
	ret

